## Copyright (c) Thomas Bock, 2016-2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(logging)) # logging

## / / / / / / / / / / / / / / / / / / / / / / / /
## Functions to run different types of analyses
## / / / / / / / / / / / / / / / / / / / / / / / /

do.analysis.c.bursts <- function(analysis.type, analysis.file.name.part, commits, analyze.files) {
  path.to.burst.data <- paste0(PATH.TO.DATA, "/", "c_bursts_", analysis.file.name.part ,"_", EPSILON, ".dat")

  all.commits <- commits

  # Remove commits on file level resp. base feature
  if(analysis.type=="function") {
    commits <- remove.file.level.commits(commits)
  } else if(analysis.type=="feature") {
    commits <- remove.base.feature.commits(commits)
  }

  if(PARALLEL.JOBS > 1 & JOB.NUMBER == "final") {
    c.bursts <- combine.parts.c.data(path.to.data=path.to.burst.data,
                                     paths.to.data.parts=paste0(PATH.TO.PART.DATA, "/", "c_bursts_", analysis.file.name.part ,"_", EPSILON, "_part", c(1:PARALLEL.JOBS), ".dat"))
  } else {
    c.bursts <- get.c.bursts(commit.data = commits,
                             path.to.c.burst.data = path.to.burst.data,
                             path.to.c.burst.data.part = paste0(PATH.TO.PART.DATA, "/", "c_bursts_", analysis.file.name.part ,"_", EPSILON, "_part", JOB.NUMBER, ".dat"),
                             analyzeFiles = analyze.files)
  }

  if(file.exists(path.to.burst.data)) {
    loginfo("Analyze c-bursts...")
    output.paths <- paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", analysis.file.name.part,"_", EPSILON)
    analyze.co.commit.bursts(c.bursts[[1]], c.bursts[[2]], commits, all.commits,
                             use.all.commits=FALSE, analyzeFiles=analyze.files,
                             output.path.dat=paste0(output.paths, ".dat"),
                             output.path.txt=paste0(output.paths, ".txt"))
    loginfo(paste0("Finished ", analysis.type, "-based c-burst analysis."))
  }
}

do.analysis.e.bursts <- function(emails, simulated=FALSE, thread.based=FALSE,
                                 based.on.commit.messages=FALSE, c.burst.artefact=NULL) {

  file.name <- "e_bursts_"
  if(thread.based) {
	file.name <- "e_bursts_thread_"
  }
  if(based.on.commit.messages) {
        file.name <- paste0(file.name, "commitMessages_", c.burst.artefact, "_")
  }

  path.to.burst.data <- paste0(PATH.TO.DATA, "/", file.name, EPSILON, ".dat")

  if(simulated) {
    # to determine e-burst of simulated time-series, the simulated email data has to be loaded
    load(paste0(PATH.TO.DATA, "/", "e_simulation.dat"))
    emails <- simulated.email.data
  }

  if(based.on.commit.messages) {
    c.burst.data <- list()

    # load the c-burst data containing the commit messages
    load(paste0(PATH.TO.DATA, "/", "c_bursts_qualitative_analysis_", c.burst.artefact, "_" , EPSILON, ".dat"))
    c.burst.data$c.bursts <- burst.commits

    # additionally, to determine the pair numbers in the c-burst data, load additional data
    load(paste0(PATH.TO.DATA, "/", "c_bursts_", c.burst.artefact, "_", EPSILON, ".dat"))
    c.burst.data$pairs.of.developers.data.frame.from.c.bursts <- bursts[[1]]
    rm(bursts)
  }

  if(PARALLEL.JOBS > 1 & JOB.NUMBER == "final") {
    e.bursts <- combine.parts.e.data(path.to.data=path.to.burst.data,
                                     paths.to.data.parts=paste0(PATH.TO.PART.DATA, "/", file.name, EPSILON, "_part", c(1:PARALLEL.JOBS), ".dat"))
  } else {
    e.bursts <- get.e.bursts(email.data = emails, simulated=simulated,
                             thread.based=thread.based,
                             path.to.e.burst.data = path.to.burst.data,
                             path.to.e.burst.data.part = paste0(PATH.TO.PART.DATA, "/", file.name, EPSILON, "_part", JOB.NUMBER, ".dat"),
                             based.on.commit.messages=based.on.commit.messages, c.burst.data=c.burst.data)
  }

  if(file.exists(path.to.burst.data)) {
    loginfo("Analyze e-bursts...")
    output.paths <- paste0(PATH.TO.DATA, "/", file.name ,"analysis_", EPSILON)
    analyze.email.bursts(e.bursts[[1]], e.bursts[[2]], emails,
                         output.path.dat=paste0(output.paths, ".dat"),
                         output.path.txt=paste0(output.paths, ".txt"), simulated=simulated)
    loginfo("Finished e-burst analysis.")
  }
}

do.code.growth.analysis <- function(analysis.type, analysis.file.name.part, commits, analyze.files) {
  path.to.burst.data <- paste0(PATH.TO.DATA, "/", "c_bursts_", analysis.file.name.part ,"_", EPSILON, ".dat")

  # Remove commits on file level resp. base feature
  if(analysis.type=="function") {
    analyzed.commits <- remove.file.level.commits(commits)
  } else if(analysis.type=="feature") {
    analyzed.commits <- remove.base.feature.commits(commits)
  } else {
    analyzed.commits <- commits
  }

  if(file.exists(path.to.burst.data)) {
    output.path <- paste0(PATH.TO.DATA, "/", "commit_sync_analysis_", analysis.file.name.part,"_", EPSILON, ".dat")

    if(!file.exists(output.path)) {
      loginfo(paste("Load c-bursts from file:", path.to.burst.data))
      load(path.to.burst.data)
      loginfo(paste("Determine synchronous and non-synchronous commits."))

      commit.hashes <- determine.synchronous.and.non.synchronous.commits(analyzed.commits, bursts[[1]],  bursts[[2]], analyze.files)
      save(commit.hashes, file=output.path)
      loginfo(paste("Saved synchronous and non-synchronous commit-hashes to file:", output.path))
    } else {
      loginfo(paste("Load synchronous and non-synchronous commit-hashes from file:", output.path))
      load(output.path)
    }

    loginfo("Calculate code growth metrics.")
    output.path.sync.txt <- paste0(PATH.TO.DATA, "/", "code_growth_metrics_sync_", analysis.file.name.part,"_", EPSILON, ".txt")
    output.path.non.sync.txt <- paste0(PATH.TO.DATA, "/", "code_growth_metrics_non_sync_", analysis.file.name.part,"_", EPSILON, ".txt")
    output.path.dat <- paste0(PATH.TO.DATA, "/", "code_growth_metrics_", analysis.file.name.part,"_", EPSILON, ".dat")

    loginfo("Read raw data added and deleted lines.")
    path.to.lines.data <- paste0(PATH.TO.DATA, "/", CASESTUDY, "_lines.csv")
    added.deleted.lines.data <- read.added.deleted.lines.data(path.to.lines.data)

    loginfo("Start code-growth calculations.")
    calculations <- list()
    calculations$synchronous.metrics <- calculate.code.growth.metrics(commit.hashes$synchronous, commits, analyzed.commits, added.deleted.lines.data, analysis.type, output.path.sync.txt)
    calculations$non.synchronous.metrics <- calculate.code.growth.metrics(commit.hashes$non.synchronous, commits, analyzed.commits, added.deleted.lines.data, analysis.type, output.path.non.sync.txt)

    save(calculations, file=output.path.dat)
    loginfo(paste("Saved code growth metrics to file:", output.path.dat))
  } else {
    loginfo(paste("File not found:", path.to.burst.data))
  }
}

get.bursts <- function(raw.data, path.to.burst.data, path.to.burst.data.part, burst.type, analyzeFiles=FALSE, simulated=FALSE, thread.based=FALSE,
                       based.on.commit.messages=FALSE, c.burst.data=NULL) {

  # If the corresponding bursts are already saved to file, load the file.
  # Otherwise determine the bursts and save them to file
  if(file.exists(path.to.burst.data)) {
    loginfo(paste("Load", burst.type, "from file:", path.to.burst.data))
    load(path.to.burst.data)
  } else {
    loginfo(paste("No", burst.type, "saved. Start determining", burst.type, "..."))

    if(burst.type=="c-bursts") {
      bursts <- determine.co.commit.bursts(raw.data, analyzeFiles)
    } else if(burst.type=="e-bursts") {
      bursts <- determine.email.bursts(raw.data, simulated, thread.based, based.on.commit.messages, c.burst.data)
    } else {
      logerror(paste("Could not determine", burst.type))
      stop()
    }
    loginfo(paste("Finished", burst.type, "identification."))
    if(PARALLEL.JOBS == 1) {
      path.to.burst.data.part <- path.to.burst.data
    }
    save(bursts, file=path.to.burst.data.part)
    loginfo(paste("Saved", burst.type, "to file:", path.to.burst.data.part))
  }
  return(bursts)
}

get.c.bursts <- function(commit.data, path.to.c.burst.data, path.to.c.burst.data.part, analyzeFiles) {
  return(get.bursts(commit.data, path.to.c.burst.data, path.to.c.burst.data.part, "c-bursts", analyzeFiles=analyzeFiles))
}

get.e.bursts <- function(email.data, path.to.e.burst.data, path.to.e.burst.data.part, simulated, thread.based, based.on.commit.messages=FALSE, c.burst.data=NULL) {
  return(get.bursts(email.data, path.to.e.burst.data, path.to.e.burst.data.part, "e-bursts", simulated=simulated, thread.based=thread.based,
                    based.on.commit.messages=based.on.commit.messages, c.burst.data=c.burst.data))
}

combine.data.frames <- function(list.of.data.frames) {
  return(rbind.fill(list.of.data.frames))
}

load.part.data <- function(paths.to.data) {
  full.data1 <- list()
  full.data2 <- list()

  for(part in paths.to.data) {
    if(!file.exists(part)) {
      stop(paste("File does not exist:", part))
    }
    load(part)
    bursts.part1 <- bursts[[1]]
    bursts.part2 <- bursts[[2]]

    full.data1 <- c(full.data1, list(bursts.part1))
    full.data2 <- c(full.data2, bursts.part2)
  }
  return(list(combine.data.frames(full.data1), full.data2))
}

combine.parts.data <- function(path.to.data, paths.to.data.parts, burst.type) {
  bursts <- load.part.data(paths.to.data.parts)
  save(bursts, file=path.to.data)
  loginfo(paste("Saved combined", burst.type, "to file:", path.to.data))
  return(bursts)
}

combine.parts.c.data <- function(path.to.data, paths.to.data.parts) {
  return(combine.parts.data(path.to.data, paths.to.data.parts, "c-bursts"))
}

combine.parts.e.data <- function(path.to.data, paths.to.data.parts) {
  return(combine.parts.data(path.to.data, paths.to.data.parts, "e-bursts"))
}

## / / / / / / / / / / / / / / / / / / / / / / / /
## Functions to do qualitative analyis
## / / / / / / / / / / / / / / / / / / / / / / / /

do.qualitative.analysis.c.e.bursts <- function(analysis.type, analysis.file.name.part, commits, analyze.files, emails, thread.based=FALSE, only.c.analysis=FALSE,
                                               based.on.commit.messages=FALSE) {

  loginfo("Start preparation for qualitative analysis...")

  ## 1) get c-burst commits

  path.to.c.burst.data <- paste0(PATH.TO.DATA, "/", "c_bursts_", analysis.file.name.part ,"_", EPSILON, ".dat")

  all.commits <- commits

  # Remove commits on file level resp. base feature
  if(analysis.type=="function") {
    commits <- remove.file.level.commits(commits)
  } else if(analysis.type=="feature") {
    commits <- remove.base.feature.commits(commits)
  }

  if(file.exists(path.to.c.burst.data)) {
    loginfo("Load c-bursts and get commits for bursts...")
    load(path.to.c.burst.data)
    c.bursts <- bursts

    loginfo("Load synchronicity data...")
    path.to.synchronicity.data <- paste0(PATH.TO.DATA, "/", "commit_sync_analysis_", analysis.file.name.part,"_", EPSILON, ".dat")

    synchronous.commit.hashes <- NULL
    if(file.exists(path.to.synchronicity.data)) {
      load(path.to.synchronicity.data)
      synchronous.commit.hashes <- commit.hashes$synchronous
    } else {
      loginfo(paste("No synchronicity data available:", path.to.synchronicity.data))
      stop()
    }

    ## get commit messages

    loginfo("Read commit messages...")

    tagging <- "proximity"
    if(analysis.type == "feature") {
      tagging <- "feature"
    }

    commit.messages <- read.commit.messages.data(paste0(PATH.TO.DATA, "/", "commitMessages_", tagging, ".list"), filter.for.first.line = TRUE)

    # get co-commit bursts

    burst.commits <- get.burst.commits(commits=commits, commit.messages=commit.messages,
                                       synchronous.commits=synchronous.commit.hashes,
                                       commit.bursts=c.bursts[[2]],
                                       pairs.of.developers.data.frame=c.bursts[[1]],
                                       analyzeFiles=analyze.files)

    output.paths <- paste0(PATH.TO.DATA, "/", "c_bursts_qualitative_analysis_", analysis.file.name.part,"_", EPSILON)
    save(burst.commits, file = paste0(output.paths,".dat"))

    loginfo(paste0("Finished ", analysis.type, "-based c-burst analysis."))
  } else {
    loginfo(paste("No c-burst data available:", path.to.burst.data))
  }

  if(!only.c.analysis) {

    ## 2) get mbox-parsing results

    loginfo("Read mbox-parsing results...")

    tagging <- "proximity_filepath"
    if(analysis.type == "feature") {
      tagging <- "feature"
    } else if(analysis.type == "function") {
      tagging <- "proximity" # "proximity_filepath"
    } else if(analysis.type == "file") {
      tagging <- "proximity_file" # "proximity_file_filepath"
    }

    path.to.mbox.parsing.data <- paste0(PATH.TO.DATA, "/", "mboxparsing_", tagging, ".list")
    mboxparsing.data <- read.mboxparsing.data(path.to.mbox.parsing.data, artefact.type = analysis.type, commit.data = commits)

    ## 3) get e-burst emails

    file.name <- "e_bursts_"
    if(thread.based) {
      file.name <- "e_bursts_thread_"
    }
    if(based.on.commit.messages) {
      file.name <- paste0(file.name, "commitMessages_", analysis.file.name.part, "_")
    }

    path.to.e.burst.data <- paste0(PATH.TO.DATA, "/", file.name, EPSILON, ".dat")

    if(file.exists(path.to.e.burst.data)) {
      loginfo("Load e-bursts and get commits for bursts...")
      load(path.to.e.burst.data)
      e.bursts <- bursts

      burst.emails <- get.burst.emails(emails=emails,
                                       email.bursts=e.bursts[[2]],
                                       pairs.of.developers.data.frame=e.bursts[[1]],
                                       mboxparsing.data=mboxparsing.data,
                                       thread.based=thread.based)

      output.paths <- paste0(PATH.TO.DATA, "/", file.name ,"mbox_", analysis.file.name.part, "_qualitative_analysis_", EPSILON)
      save(burst.emails, file = paste0(output.paths,".dat"))
      loginfo("Finished e-burst analysis.")
    } else {
      loginfo(paste("No e-burst data available:", path.to.burst.data))
    }

    ## 4) check for developer pairs which have bursts

    pair.data <- get.developer.pairs.having.c.and.e.bursts(c.bursts[[1]], burst.commits, e.bursts[[1]], burst.emails)

    output.paths.developer.pairs.c <- paste0(PATH.TO.DATA, "/", analysis.file.name.part, "_qualitative_analysis_", EPSILON)
    output.paths.developer.pairs.e <- paste0(PATH.TO.DATA, "/", file.name, "qualitative_analysis_", EPSILON)
    output.paths.developer.pairs.ce <- paste0(PATH.TO.DATA, "/", file.name , analysis.file.name.part, "_qualitative_analysis_", EPSILON)

    developer.pairs.c.bursts <- pair.data[[1]]
    developer.pairs.e.bursts <- pair.data[[2]]
    developer.pairs.c.and.e.bursts <- pair.data[[3]]

    save(developer.pairs.c.bursts, file=paste0(output.paths.developer.pairs.c, "_devCbursts.dat"))
    save(developer.pairs.e.bursts, file=paste0(output.paths.developer.pairs.e, "_devEbursts.dat"))
    save(developer.pairs.c.and.e.bursts, file=paste0(output.paths.developer.pairs.ce, "_devCEbursts.dat"))

  }
}

## / / / / / / / / / / / / / / / / / / / / / / / /
## Functions to compute response times
## / / / / / / / / / / / / / / / / / / / / / / / /

compute.response.times <- function(commits.or.emails, analysis.type, raw.data, analyze.files) {
  if(commits.or.emails=="commits") {
    compute.response.times.commits(analysis.type, raw.data, analyze.files = analyze.files)
  } else if(commits.or.emails=="emails") {
    compute.response.times.emails(analysis.type, raw.data)
  } else {
    logerror(paste("Could not determine", commits.or.emails))
    stop()
  }
}

compute.response.times.commits <- function(analysis.type, commits, analyze.files = FALSE){

  # Remove commits on file level resp. base feature
  if(analysis.type=="function") {
    commits <- remove.file.level.commits(commits)
  } else if(analysis.type=="feature") {
    commits <- remove.base.feature.commits(commits)
  }

  output.path.dat <- paste0(PATH.TO.DATA, "/", "response_times_commits_", analysis.type, ".dat")
  determine.response.times.in.commits.on.artifacts(commits, analyzeFiles=FALSE, output.path.dat)
}

compute.response.times.emails <- function(analysis.type, emails){

  if(analysis.type=="threads") {
    output.path.dat <- paste0(PATH.TO.DATA, "/", "response_times_emails_", analysis.type, ".dat")
    determine.response.times.in.threads(emails, output.path.dat)
  } else {
    output.path.dat <- paste0(PATH.TO.DATA, "/", "response_times_emails_", analysis.type, ".dat")
    determine.response.times.in.emails(emails, output.path.dat)
  }
}
