## Copyright (c) Thomas Bock, 2016-2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(parallel)) # tools for parallel execution

## / / / / / / / / / / / / / / / / / / / / / /
## Code-Growth Metrics in Co-Commit Bursts
## / / / / / / / / / / / / / / / / / / / / / /

# Determines synchronous and non-synchronous commits and calculates code-growth metrics for each commit
determine.synchronous.and.non.synchronous.commits <- function(commits, pairs.of.developers, commit.bursts, analyzeFiles=FALSE) {

  # check whether this analysis is file-based or not
  artefact.column <- get.artefact.column(analyzeFiles)

  # Iterate over commits
  commit.hashes <- unique(commits[["commitHash"]])

  synchronous.and.non.synchronous.commits <- mclapply(commit.hashes, mc.cores=CORES, FUN=function(commit){
    synchronous.commits <- list()
    non.synchronous.commits <- list()

    author.of.commit <- unique(commits[commits["commitHash"]==commit,][["author"]])
    time.stamp.of.commit <- unique(commits[commits["commitHash"]==commit,][["authorDate"]])
    artefacts.of.commit <- unique(commits[commits["commitHash"]==commit,][[artefact.column]])

    # Get all pairs in which the author of the current commit is part of
    pairs.containing.author.data.frame <- get.pairs.for.one.developer(author.of.commit, pairs.of.developers)
    pairs.containing.author <- as.list(as.data.frame(t(pairs.containing.author.data.frame)))

    synchronous <- FALSE

    # Iterate over all pairs which the author of the current commit is part of, until a synchronous development on this commit is found
    for(pair in pairs.containing.author) {
      pair.no <- get.pair.no(pair[1], pair[2], pairs.containing.author.data.frame)
      bursts.data.frame <- commit.bursts[[pair.no]]

      # Only go on with this pair if there is at least one burst for the current pair
      if(length(bursts.data.frame)>0) {
        burst.containing.commit <- bursts.data.frame[(ymd_hms(bursts.data.frame[["start"]]) <= ymd_hms(time.stamp.of.commit)
                                                      & ymd_hms(bursts.data.frame[["end"]]) >= ymd_hms(time.stamp.of.commit)),]

        # Check if there is burst of the current pair which timely contains the current commit
        if(nrow(burst.containing.commit) > 0) {
          # Get the commits of the other author of the pair within the time-interval of the current burst
          partner <- if(pair[1]==author.of.commit) pair[2] else pair[1]
          commits.of.partner <- commits[(commits["author"]==partner & ymd_hms(commits[["authorDate"]]) >= ymd_hms(burst.containing.commit[["start"]])
                                         & ymd_hms(commits[["authorDate"]]) <= ymd_hms(burst.containing.commit[["end"]])),]

          # Get the artefacts the other developer committed to within the burst time-interval and check for overlapping artefacts with the current commit
          artefacts.commited.to.by.partner <- unique(commits.of.partner[[artefact.column]])
          artefacts.commited.to.by.both <- any(artefacts.of.commit %in% artefacts.commited.to.by.partner)

          if(artefacts.commited.to.by.both) {
            # commit is synchronous
            synchronous <- TRUE
            break
          }
        }
      }
    }

    return(synchronous)
  })

  synchronous.commits <- commit.hashes[which(synchronous.and.non.synchronous.commits==TRUE)]
  non.synchronous.commits <- commit.hashes[which(synchronous.and.non.synchronous.commits==FALSE)]

  commit.hashes <- list()
  commit.hashes$synchronous <- synchronous.commits
  commit.hashes$non.synchronous <- non.synchronous.commits

  return(commit.hashes)
}

calculate.code.growth.metrics <- function(commit.hashes.to.consider, all.commits, analyzed.commits, added.deleted.lines, analysis.type, output.path.txt) {

  # for each commit hash in the given list of commit hashes compute several code growth metrics
  code.growth.metrics.per.commit <- mclapply(commit.hashes.to.consider, mc.cores=CORES, FUN=function(commitHash){

    # get all information of the analyzed commit
    analyzed.commit <- analyzed.commits[analyzed.commits[,"commitHash"]==commitHash,]

    # get all information of the analyzed commit inclusive base feature commits and file level commits (which might be removed for burst analysis)
    commit.incl.removed.artefacts <- all.commits[all.commits[,"commitHash"]==commitHash,]

    growth <- list()
    growth$added <- get.added.lines.of.commit(analyzed.commit, analysis.type, commit.incl.removed.artefacts, added.deleted.lines)
    growth$deleted <- get.deleted.lines.of.commit(analyzed.commit, analysis.type, commit.incl.removed.artefacts, added.deleted.lines)

    growth$normalizedAdded <- get.added.lines.of.commit(analyzed.commit, analysis.type, commit.incl.removed.artefacts, added.deleted.lines,  normalize=TRUE)
    growth$normalizedDeleted <- get.deleted.lines.of.commit(analyzed.commit, analysis.type, commit.incl.removed.artefacts, added.deleted.lines, normalize=TRUE)

    return(growth)
  })

  # create a single vector for each metric containing the results for all commits
  added.lines.per.commit <- unlist(mclapply(as.list(1:length(commit.hashes.to.consider)), mc.cores=CORES, FUN=function(i, code.growth.metrics.per.commit){
    return(code.growth.metrics.per.commit[[i]]$added)}, code.growth.metrics.per.commit=code.growth.metrics.per.commit))
  deleted.lines.per.commit <- unlist(mclapply(as.list(1:length(commit.hashes.to.consider)), mc.cores=CORES, FUN=function(i, code.growth.metrics.per.commit){
    return(code.growth.metrics.per.commit[[i]]$deleted)}, code.growth.metrics.per.commit=code.growth.metrics.per.commit))

  normalized.added.lines.per.commit <- unlist(mclapply(as.list(1:length(commit.hashes.to.consider)), mc.cores=CORES, FUN=function(i, code.growth.metrics.per.commit){
    return(code.growth.metrics.per.commit[[i]]$normalizedAdded)}, code.growth.metrics.per.commit=code.growth.metrics.per.commit))
  normalized.deleted.lines.per.commit <- unlist(mclapply(as.list(1:length(commit.hashes.to.consider)), mc.cores=CORES, FUN=function(i, code.growth.metrics.per.commit){
    return(code.growth.metrics.per.commit[[i]]$normalizedDeleted)}, code.growth.metrics.per.commit=code.growth.metrics.per.commit))

  metrics <- list()
  metrics$added <- added.lines.per.commit
  metrics$deleted <- deleted.lines.per.commit
  metrics$growth <- added.lines.per.commit - deleted.lines.per.commit
  metrics$effort <- added.lines.per.commit + deleted.lines.per.commit
  metrics$normalizedAdded <- normalized.added.lines.per.commit
  metrics$normalizedDeleted <- normalized.deleted.lines.per.commit
  metrics$normalizedGrowth <- normalized.added.lines.per.commit - normalized.deleted.lines.per.commit
  metrics$normalizedEffort <- normalized.added.lines.per.commit + normalized.deleted.lines.per.commit

  # Generate textual output for all determined results (from all the lists)

  output.number.of.commits<- paste("Number of commits:", length(commit.hashes.to.consider), sep=" ")

  output.median.added <- paste("Number of added lines per commit: (median)", median(metrics$added), sep=" ")
  output.mean.added <- paste("Number of added lines per commit: (mean)", mean(metrics$added), sep=" ")
  output.min.added <- paste("Number of added lines per commit: (min)", min(metrics$added), sep=" ")
  output.max.added <- paste("Number of added lines per commit: (max)", max(metrics$added), sep=" ")

  output.median.deleted <- paste("Number of deleted lines per commit: (median)", median(metrics$deleted), sep=" ")
  output.mean.deleted <- paste("Number of deleted lines per commit: (mean)", mean(metrics$deleted), sep=" ")
  output.min.deleted <- paste("Number of deleted lines per commit: (min)", min(metrics$deleted), sep=" ")
  output.max.deleted <- paste("Number of deleted lines per commit: (max)", max(metrics$deleted), sep=" ")

  output.median.growth <- paste("Code growth per commit: (median)", median(metrics$growth), sep=" ")
  output.mean.growth <- paste("Code growth per commit: (mean)", mean(metrics$growth), sep=" ")
  output.min.growth <- paste("Code growth per commit: (min)", min(metrics$growth), sep=" ")
  output.max.growth <- paste("Code growth per commit: (max)", max(metrics$growth), sep=" ")

  output.median.effort <- paste("Code effort per commit: (median)", median(metrics$effort), sep=" ")
  output.mean.effort <- paste("Code effort per commit: (mean)", mean(metrics$effort), sep=" ")
  output.min.effort <- paste("Code effort per commit: (min)", min(metrics$effort), sep=" ")
  output.max.effort <- paste("Code effort per commit: (max)", max(metrics$effort), sep=" ")

  output.median.normalized.added <- paste("Normalized number of added lines per commit: (median)", median(metrics$normalizedAdded), sep=" ")
  output.mean.normalized.added <- paste("Normalized number of added lines per commit: (mean)", mean(metrics$normalizedAdded), sep=" ")
  output.min.normalized.added <- paste("Normalized number of added lines per commit: (min)", min(metrics$normalizedAdded), sep=" ")
  output.max.normalized.added <- paste("Normalized number of added lines per commit: (max)", max(metrics$normalizedAdded), sep=" ")

  output.median.normalized.deleted <- paste("Normalized number of deleted lines per commit: (median)", median(metrics$normalizedDeleted), sep=" ")
  output.mean.normalized.deleted <- paste("Normalized number of deleted lines per commit: (mean)", mean(metrics$normalizedDeleted), sep=" ")
  output.min.normalized.deleted <- paste("Normalized number of deleted lines per commit: (min)", min(metrics$normalizedDeleted), sep=" ")
  output.max.normalized.deleted <- paste("Normalized number of deleted lines per commit: (max)", max(metrics$normalizedDeleted), sep=" ")

  output.median.normalized.growth <- paste("Normalized code growth per commit: (median)", median(metrics$normalizedGrowth), sep=" ")
  output.mean.normalized.growth <- paste("Normalized code growth per commit: (mean)", mean(metrics$normalizedGrowth), sep=" ")
  output.min.normalized.growth <- paste("Normalized code growth per commit: (min)", min(metrics$normalizedGrowth), sep=" ")
  output.max.normalized.growth <- paste("Normalized code growth per commit: (max)", max(metrics$normalizedGrowth), sep=" ")

  output.median.normalized.effort <- paste("Normalized code effort per commit: (median)", median(metrics$normalizedEffort), sep=" ")
  output.mean.normalized.effort <- paste("Normalized code effort per commit: (mean)", mean(metrics$normalizedEffort), sep=" ")
  output.min.normalized.effort <- paste("Normalized code effort per commit: (min)", min(metrics$normalizedEffort), sep=" ")
  output.max.normalized.effort <- paste("Normalized code effort per commit: (max)", max(metrics$normalizedEffort), sep=" ")

  output <- paste(output.number.of.commits,
                  output.median.added, output.mean.added, output.min.added, output.max.added,
                  output.median.deleted, output.mean.deleted, output.min.deleted, output.max.deleted,
                  output.median.growth, output.mean.growth, output.min.growth, output.max.growth,
                  output.median.effort, output.mean.effort, output.min.effort, output.max.effort,
                  output.median.normalized.added, output.mean.normalized.added, output.min.normalized.added, output.max.normalized.added,
                  output.median.normalized.deleted, output.mean.normalized.deleted, output.min.normalized.deleted, output.max.normalized.deleted,
                  output.median.normalized.growth, output.mean.normalized.growth, output.min.normalized.growth, output.max.normalized.growth,
                  output.median.normalized.effort, output.mean.normalized.effort, output.min.normalized.effort, output.max.normalized.effort, "",
                  sep="\n")
  cat(output)
  write(output, file=output.path.txt)

  return(metrics)
}

get.added.lines.of.commit <- function(analyzed.commit, analysis.type, commit.incl.removed.artefacts, added.deleted.lines, normalize=FALSE) {
  added <- analyzed.commit[1, "addedLines"]

  added.non.header <- sum(analyzed.commit[["size"]])

  added <- added.non.header

  # sanity check:
  # extract added lines in this commit for the analyzed files
  analyzed.files <- unique(analyzed.commit[, "file"])
  added.lines.of.analyzed.commit <- added.deleted.lines[added.deleted.lines[,"commitHash"]==analyzed.commit[1, "commitHash"],]
  added.lines.of.analyzed.files <- added.lines.of.analyzed.commit[added.lines.of.analyzed.commit[,"file"] %in% analyzed.files, "addedLines"]
  added2 <- sum(added.lines.of.analyzed.files)
  if(added != added2) print(paste("Commit:", unique(analyzed.commit[["commitHash"]]), "Codeface:", added, "GitShowNumstat:", added2))

  # Use 'git show --numstat information'
  added <- added2

  if(analysis.type=="feature") {
    # currently not needed
  } else if(analysis.type=="function") {
    # currently not needed
  }

  if(normalize) {
    # normalize by the number of changed artefacts within the analyzed commit

    if(analysis.type=="feature" || analysis.type=="function") {
      artefacts <- length(unique(analyzed.commit[["artefact"]]))
    } else if(analysis.type=="file") {
      artefacts <- length(unique(analyzed.commit[["file"]]))
    }
    added <- added / artefacts
  }
  return(added)
}

get.deleted.lines.of.commit <- function(analyzed.commit, analysis.type, commit.incl.removed.artefacts, added.deleted.lines, normalize=FALSE) {
  deleted <- analyzed.commit[1, "deletedLines"]

  # This is just a guessed heuristic, to determine the deleted lines of unconsidered artefacts. Actually, that might be completely wrong!
  # added <- analyzed.commit[1, "addedLines"]
  # added.base <- added - sum(analyzed.commit[["size"]])
  # deleted <- max(0, deleted - (0.5 * added.base))

  # new approach: consider ratio between added and deleted lines of commit
  #added.all <- analyzed.commit[1, "addedLines"]
  #added.after.removing.unconsidered.artifacts <- sum(analyzed.commit[["size"]])
  #added.ratio <- added.after.removing.unconsidered.artifacts / added.all
  #deleted.ratio <- added.ratio
  #deleted.after.removing.unconsidered.artifacts <- deleted * deleted.ratio

  # extract deleted lines in this commit for the analyzed files
  analyzed.files <- unique(analyzed.commit[, "file"])
  deleted.lines.of.analyzed.commit <- added.deleted.lines[added.deleted.lines[,"commitHash"]==analyzed.commit[1, "commitHash"],]
  deleted.lines.of.analyzed.files <- deleted.lines.of.analyzed.commit[deleted.lines.of.analyzed.commit[,"file"] %in% analyzed.files, "deletedLines"]
  deleted.after.removing.unconsidered.artifacts <- sum(deleted.lines.of.analyzed.files)

  if(normalize) {
    # normalize by the number of changed artefacts within the analyzed commit

    if(analysis.type=="feature" || analysis.type=="function") {
      artefacts <- length(unique(analyzed.commit[["artefact"]]))
    } else if(analysis.type=="file") {
      artefacts <- length(unique(analyzed.commit[["file"]]))
    }
    deleted.after.removing.unconsidered.artifacts <- deleted.after.removing.unconsidered.artifacts / artefacts
  }
  return(deleted.after.removing.unconsidered.artifacts)
}
