## Copyright (c) Thomas Bock, 2016-2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(parallel)) # tools for parallel execution

## / / / / / / / / / / / / / /
## Co-Commit Bursts
## / / / / / / / / / / / / / /

# Determines co-commit bursts.
# Returns a list which first element is the data frame of developer pairs
# and which second element is a list containing data frames of bursts for each developer pair
determine.co.commit.bursts <- function(commits, analyzeFiles=FALSE) {

  # check whether this analysis is file-based or not
  artefact.column <- get.artefact.column(analyzeFiles)

  developers <- unique(commits[,"author"])

  pairs.of.developers.data.frame <- as.data.frame(t(combn(developers, 2, simplify=TRUE)))
  colnames(pairs.of.developers.data.frame) <- c("d_a", "d_b")

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))
  # pairs.of.developers <- split(pairs.of.developers.data.frame, seq(1, nrow(pairs.of.developers.data.frame)))

  if(PARALLEL.JOBS > 1) {
    pairs.of.developers.data.frame <- split(pairs.of.developers.data.frame, cut(seq(1,length(pairs.of.developers)), PARALLEL.JOBS))[[JOB.NUMBER]]
    pairs.of.developers <- split(pairs.of.developers, cut(seq(1,length(pairs.of.developers)), PARALLEL.JOBS))[[JOB.NUMBER]]
    loginfo(paste("Developer pairs to analyze in this job: ", length(pairs.of.developers)))
  }

  # apply burst detection to each pair of developers
  list.of.c.bursts <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    commits.d_a <- commits[commits["author"] == pair[1],]
    commits.d_b <- commits[commits["author"] == pair[2],]

    commits.d_a.both.commit.to.artefact <- commits.d_a[commits.d_a[,artefact.column] %in% unique(commits.d_b[,artefact.column]),]

    commit.bursts.a.b <- list()

    # check whether the two developers work on the same artefact anytime. If not, there can not be a burst between them.
    if(nrow(commits.d_a.both.commit.to.artefact) > 0) {
      commits.d_b.both.commit.to.artefact <- (commits.d_b[commits.d_b[,artefact.column] %in% unique(commits.d_a.both.commit.to.artefact[,artefact.column]),])
      commits.d_a.both.commit.to.artefact.list <- split(commits.d_a.both.commit.to.artefact, seq(1,nrow(commits.d_a.both.commit.to.artefact)))

      lapply(commits.d_a.both.commit.to.artefact.list, epsilon=EPSILON, FUN=function(c.d_a, epsilon) {
        c.d_a[["authorDate"]] <- sapply(c.d_a[["authorDate"]], FUN=function(x) fix.midnight(x), USE.NAMES=FALSE)
        commits.d_b.both.commit.to.artefact[["authorDate"]] <- sapply(commits.d_b.both.commit.to.artefact[["authorDate"]], FUN=function(x) fix.midnight(x), USE.NAMES=FALSE)


        commit.date.d_a <- ymd_hms(c.d_a["authorDate"])

        # select commits of d_b which were committed at commit.date.d_a +/- epsilon days
        commits.d_b.same.time <- commits.d_b.both.commit.to.artefact[abs(difftime(commit.date.d_a, ymd_hms(commits.d_b.both.commit.to.artefact[,"authorDate"]))) <= duration(epsilon, "day"),]

        # check if there actually is at least one commit of d_b committed at commit.date.d_a +/- epsilon days
        if(nrow(commits.d_b.same.time) > 0) {

          # select the commits of d_b in the specified time interval that touch at least one same artefact of the commit of d_a
          commits.d_b.same.artefact <- commits.d_b.same.time[commits.d_b.same.time[,artefact.column] %in% unique(c.d_a[,artefact.column]),]

          # check if there actually is a commit burst. Only if there is a burst, add the burst to the list of bursts
          if(nrow(commits.d_b.same.artefact) > 0) {

            # determine beginning and end of commit burst
            new.commit.burst.start <- min(commit.date.d_a, ymd_hms(commits.d_b.same.artefact[,"authorDate"]))
            new.commit.burst.end <- max(commit.date.d_a, ymd_hms(commits.d_b.same.artefact[,"authorDate"]))
            new.commit.burst <- data.frame(start=new.commit.burst.start,
                                           end=new.commit.burst.end,
                                           occurence=new.commit.burst.start+(new.commit.burst.end-new.commit.burst.start)/2)

            new.commit.burst$start <- fix.midnight(new.commit.burst$start)
            new.commit.burst$end <- fix.midnight(new.commit.burst$end)
            new.commit.burst$occurence <- fix.midnight(new.commit.burst$occurence)

            # first check if commit burst overlaps with some other commit burst
            l <- length(commit.bursts.a.b)
            bursts.to.remove <- c()

            if(l > 0) {
              for(i in c(1:length(commit.bursts.a.b))) {
                burst <- commit.bursts.a.b[i][[1]]

                # only check for overlaps if some preconditions are fulfilled
                if(year(burst$end) >= year(new.commit.burst$start) & year(burst$start) <= year(new.commit.burst.end)) {

                  burst.start <-ymd_hms(burst$start)
                  burst.end <-ymd_hms(burst$end)
                  new.commit.burst.start <-ymd_hms(new.commit.burst$start)
                  new.commit.burst.end <-ymd_hms(new.commit.burst$end)

                  if(int_overlaps(
                    interval(burst.start, burst.end),
                    interval(new.commit.burst.start, new.commit.burst.end)
                  )) {

                    new.commit.burst.start <- min(burst.start, new.commit.burst.start)
                    new.commit.burst.end <- max(burst.end, new.commit.burst.end)
                    new.commit.burst <- data.frame(start=new.commit.burst.start,
                                                   end=new.commit.burst.end,
                                                   occurence=new.commit.burst.start+(new.commit.burst.end-new.commit.burst.start)/2)

                    new.commit.burst$start <- fix.midnight(new.commit.burst$start)
                    new.commit.burst$end <- fix.midnight(new.commit.burst$end)
                    new.commit.burst$occurence <- fix.midnight(new.commit.burst$occurence)

                    bursts.to.remove <- c(bursts.to.remove, i)
                  }
                }
              }
            }

            # add new commit burst (or merged one) to list of bursts for the current pair of developers
            if(length(bursts.to.remove) > 0) {
              commit.bursts.a.b <<- c(commit.bursts.a.b[-bursts.to.remove], list(new.commit.burst))
            } else {
              commit.bursts.a.b <<- c(commit.bursts.a.b, list(new.commit.burst))
            }
          }
        }
      } )

      # combine all list elements to one data frame
      commit.bursts.a.b <- rbind.fill(commit.bursts.a.b)
    }
    return(commit.bursts.a.b)
  })

  #  for(x in rownames(pairs.of.developers.data.frame)) {
  #    print(pairs.of.developers.data.frame[x,])
  #    print(map.of.c.bursts[[x]])
  #  }

  # return [1] data frame containing the pairs of developers
  # return [2] list which contains a data frame of bursts for each developer pair
  return(list(pairs.of.developers.data.frame,list.of.c.bursts))
}

# Analyzes number of bursts per pair, length of bursts, and calculates several kinds of synchronicity degrees
# Results are printed to console
analyze.co.commit.bursts <- function(pairs.of.developers.data.frame, list.of.c.bursts, commits, all.commits, use.all.commits, analyzeFiles=FALSE, output.path.dat, output.path.txt) {

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  # get overall number of bursts
  number.bursts.overall <- get.number.of.bursts.overall(list.of.c.bursts)

  # iterate over all pairs to determine the number of c-bursts per pair
  number.of.intervals <- get.number.of.bursts.per.pair(pairs.of.developers, pairs.of.developers.data.frame, list.of.c.bursts)

  # iterate over all c-bursts for each pair to determine the lengths of bursts
  length.of.intervals <- get.lengths.of.bursts(pairs.of.developers, pairs.of.developers.data.frame, list.of.c.bursts)

  # iterate over all c-bursts to determine the number of commit activities per burst
  commit.activities.per.burst <- get.activities.of.bursts(pairs.of.developers, pairs.of.developers.data.frame, list.of.c.bursts, "commit", commits, "authorDate")

  # iterate over all c-bursts for each pair to determine artefact-based synchronicity degree
  sync.degrees.artefacts <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    sync.degrees.artefacts <- c()
    pair.no <- get.pair.no(pair[1], pair[2], pairs.of.developers.data.frame)
    bursts.data.frame <- list.of.c.bursts[[pair.no]]

    # Check if there is no burst for the current developer pair
    if(length(bursts.data.frame)>0) {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(interval in bursts) {
        ival <- interval(interval[1], interval[2])

        # Get the commits of both developers within current interval
        commits.d_a <- commits[commits[,"author"] == pair[1] & ymd_hms(commits[,"authorDate"]) %within% ival,]
        commits.d_b <- commits[commits[,"author"] == pair[2] & ymd_hms(commits[,"authorDate"]) %within% ival,]

        # Calculate synchronicity degrees (the first return value is based an artfacts only, the second one is based on LOC within artefacts)
        sync.degrees <- calculate.synchronicity.degrees(commits.d_a, commits.d_b, analyzeFiles)

        # Add calculated synchronicity degrees to the respective lists
        sync.degrees.artefacts <- c(sync.degrees.artefacts, sync.degrees[1])
        # sync.degrees.loc <- c(sync.degrees.loc, sync.degrees[2])
      }
    }
    return(sync.degrees.artefacts)
  })
  sync.degrees.artefacts <- unlist(sync.degrees.artefacts)

  # iterate over all c-bursts for each pair to determine LOC-based synchronicity degree
  sync.degrees.loc <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    sync.degrees.loc <- c()
    pair.no <- get.pair.no(pair[1], pair[2], pairs.of.developers.data.frame)
    bursts.data.frame <- list.of.c.bursts[[pair.no]]

    # Check if there is no burst for the current developer pair
    if(length(bursts.data.frame)>0) {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(interval in bursts) {
        ival <- interval(interval[1], interval[2])

        # Get the commits of both developers within current interval
        if(use.all.commits) {
          commits.d_a <- all.commits[all.commits[,"author"] == pair[1] & ymd_hms(all.commits[,"authorDate"]) %within% ival,]
          commits.d_b <- all.commits[all.commits[,"author"] == pair[2] & ymd_hms(all.commits[,"authorDate"]) %within% ival,]
        } else {
          commits.d_a <- commits[commits[,"author"] == pair[1] & ymd_hms(commits[,"authorDate"]) %within% ival,]
          commits.d_b <- commits[commits[,"author"] == pair[2] & ymd_hms(commits[,"authorDate"]) %within% ival,]
        }

        # Calculate synchronicity degrees (the first return value is based an artfacts only, the second one is based on LOC within artefacts)
        sync.degrees <- calculate.synchronicity.degrees(commits.d_a, commits.d_b, analyzeFiles)

        # Add calculated synchronicity degrees to the respective lists
        #sync.degrees.artefacts <- c(sync.degrees.artefacts, sync.degrees[1])
        sync.degrees.loc <- c(sync.degrees.loc, sync.degrees[2])
      }
    }
    return(sync.degrees.loc)
  })
  sync.degrees.loc <- unlist(sync.degrees.loc)

  # iterate over all c-bursts for each pair to determine weighted synchronicity degree using the geometric mean
  sync.degrees.weighted <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    sync.degrees.weighted <- c()
    pair.no <- get.pair.no(pair[1], pair[2], pairs.of.developers.data.frame)
    bursts.data.frame <- list.of.c.bursts[[pair.no]]

    # Check if there is no burst for the current developer pair
    if(length(bursts.data.frame)>0) {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(interval in bursts) {
        ival <- interval(interval[1], interval[2])

        # Get the commits of both developers within current interval
        if(use.all.commits) {
          commits.d_a <- all.commits[all.commits[,"author"] == pair[1] & ymd_hms(all.commits[,"authorDate"]) %within% ival,]
          commits.d_b <- all.commits[all.commits[,"author"] == pair[2] & ymd_hms(all.commits[,"authorDate"]) %within% ival,]
        } else {
          commits.d_a <- commits[commits[,"author"] == pair[1] & ymd_hms(commits[,"authorDate"]) %within% ival,]
          commits.d_b <- commits[commits[,"author"] == pair[2] & ymd_hms(commits[,"authorDate"]) %within% ival,]
        }

        # Calculate the weighted synchronicity degrees using the geometric mean
        sync.degree <- calculate.weighted.synchronicity.degree(commits.d_a, commits.d_b, analyzeFiles)

        # Add calculated synchronicity degrees to the respective lists
        #sync.degrees.artefacts <- c(sync.degrees.artefacts, sync.degrees[1])
        sync.degrees.weighted <- c(sync.degrees.weighted, sync.degree)
      }
    }
    return(sync.degrees.weighted)
  })
  sync.degrees.weighted <- unlist(sync.degrees.weighted)

  # Save calculated numbers to file
  data <- list()
  data$pairs <- pairs.of.developers.data.frame
  data$bursts.per.pair <- number.of.intervals
  data$length.of.bursts <- length.of.intervals
  data$activities.per.burst <- commit.activities.per.burst
  data$sync.degrees.artefacts <- sync.degrees.artefacts
  data$sync.degrees.loc <- sync.degrees.loc
  data$sync.degrees.weighted <- sync.degrees.weighted

  save(data, file=output.path.dat)

  # Generate textual output for all determined results (from all the lists)

  output.number.of.authors <- paste("Number of authors of commits:", length(unique(commits[,"author"])), sep=" ")
  output.number.of.bursts.overall <- paste("Number of bursts (overall):", number.bursts.overall, sep=" ")

  output.median.number.of.bursts.per.pair <- paste("Number of c-bursts per pair: (median)", median(number.of.intervals), sep=" ")
  output.mean.number.of.bursts.per.pair <- paste("Number of c-bursts per pair: (mean)", mean(number.of.intervals), sep=" ")
  output.min.number.of.bursts.per.pair <- paste("Number of c-bursts per pair: (min)", min(number.of.intervals), sep=" ")
  output.max.number.of.bursts.per.pair <- paste("Number of c-bursts per pair: (max)", max(number.of.intervals), sep=" ")

  output.median.length.of.bursts <- paste("Length of c-bursts: (median)", median(length.of.intervals), sep=" ")
  output.mean.length.of.bursts <- paste("Length of c-bursts: (mean)", mean(length.of.intervals), sep=" ")
  output.min.length.of.bursts <- paste("Length of c-bursts: (min)", min(length.of.intervals), sep=" ")
  output.max.length.of.bursts <- paste("Length of c-bursts: (max)", max(length.of.intervals), sep=" ")

  output.median.commit.activities.per.burst <- paste("Commit activities per burst: (median)", median(commit.activities.per.burst), sep=" ")
  output.mean.commit.activities.per.burst <- paste("Commit activities per burst: (mean)", mean(commit.activities.per.burst), sep=" ")
  output.min.commit.activities.per.burst <- paste("Commit activities per burst: (min)", min(commit.activities.per.burst), sep=" ")
  output.max.commit.activities.per.burst <- paste("Commit activities per burst: (max)", max(commit.activities.per.burst), sep=" ")

  output.median.sync.degree.artefacts <- paste("Artefact-based synchronicity degree: (median)", median(sync.degrees.artefacts), sep=" ")
  output.mean.sync.degree.artefacts <- paste("Artefact-based synchronicity degree: (mean)", mean(sync.degrees.artefacts), sep=" ")
  output.min.sync.degree.artefacts <- paste("Artefact-based synchronicity degree: (min)", min(sync.degrees.artefacts), sep=" ")
  output.max.sync.degree.artefacts <- paste("Artefact-based synchronicity degree: (max)", max(sync.degrees.artefacts), sep=" ")

  output.median.sync.degree.loc <- paste("LOC-based synchronicity degree: (median)", median(sync.degrees.loc), sep=" ")
  output.mean.sync.degree.loc <- paste("LOC-based synchronicity degree: (mean)", mean(sync.degrees.loc), sep=" ")
  output.min.sync.degree.loc <- paste("LOC-based synchronicity degree: (min)", min(sync.degrees.loc), sep=" ")
  output.max.sync.degree.loc <- paste("LOC-based synchronicity degree: (max)", max(sync.degrees.loc), sep=" ")

  output.median.sync.degree.weighted <- paste("Weighted LOC-based synchronicity degree: (median)", median(sync.degrees.weighted), sep=" ")
  output.mean.sync.degree.weighted <- paste("Weighted LOC-based synchronicity degree: (mean)", mean(sync.degrees.weighted), sep=" ")
  output.min.sync.degree.weighted <- paste("Weighted LOC-based synchronicity degree: (min)", min(sync.degrees.weighted), sep=" ")
  output.max.sync.degree.weighted <- paste("Weighted LOC-based synchronicity degree: (max)", max(sync.degrees.weighted), sep=" ")


  output <- paste(output.number.of.authors, output.number.of.bursts.overall,
                  output.median.number.of.bursts.per.pair, output.mean.number.of.bursts.per.pair, output.min.number.of.bursts.per.pair, output.max.number.of.bursts.per.pair,
                  output.median.length.of.bursts, output.mean.length.of.bursts, output.min.length.of.bursts, output.max.length.of.bursts,
                  output.median.commit.activities.per.burst, output.mean.commit.activities.per.burst, output.min.commit.activities.per.burst, output.max.commit.activities.per.burst,
                  output.median.sync.degree.artefacts, output.mean.sync.degree.artefacts, output.min.sync.degree.artefacts, output.max.sync.degree.artefacts,
                  output.median.sync.degree.loc, output.mean.sync.degree.loc, output.min.sync.degree.loc, output.max.sync.degree.loc,
                  output.median.sync.degree.weighted, output.mean.sync.degree.weighted, output.min.sync.degree.weighted, output.max.sync.degree.weighted, "",
                  sep="\n")
  cat(output)
  write(output, file=output.path.txt)
}

# Calculate several synchronicity degrees for the commits of two developers within a burst
# Param: commits.d_a - the commits of developer a within the considered burst
# Param: commits.d_b - the commits of developer b within the same considered burst
# Param: analyzeFiles - whether to analyze files or not (then artefacts proximity or feature are analyzed)
# Returns: [1] synchronicity degree based on artefacts only
# Returns: [2] synchronicity degree based on changed lines within artefacts
calculate.synchronicity.degrees <- function(commits.d_a, commits.d_b, analyzeFiles) {

  # check whether this analysis is file-based or not
  artefact <- get.artefact.column(analyzeFiles)

  # get changed artefacts of d_a resp. d_b
  changed.artefacts.d_a <- unique(commits.d_a[,artefact])
  changed.artefacts.d_b <- unique(commits.d_b[,artefact])

  # get changed artefacts overall
  changed.artefacts <- unique(c(changed.artefacts.d_a, changed.artefacts.d_b))

  # get changed artefacts both
  changed.artefacts.both <- intersect(changed.artefacts.d_a, changed.artefacts.d_b)

  # do not consider "Base_Feature" or "File_Level" as commonly changed artefacts
  changed.artefacts.both <- changed.artefacts.both[!changed.artefacts.both %in% c("Base_Feature", "File_Level")]

  # calc synchronicity degree based on artefacts only
  sync.degree.artefacts <- length(changed.artefacts.both)/length(changed.artefacts)


  # get changed lines of d_a resp. d_b
  sizes.changed.d_a <- sum(commits.d_a[,"size"])
  sizes.changed.d_b <- sum(commits.d_b[,"size"])

  # get changed lines overall
  sizes.changed <- sizes.changed.d_a + sizes.changed.d_b

  # get changed lines of artefacts changed by both
  sizes.changed.both.d_a <- sum(commits.d_a[commits.d_a[,artefact] %in% changed.artefacts.both,"size"])
  sizes.changed.both.d_b <- sum(commits.d_b[commits.d_b[,artefact] %in% changed.artefacts.both,"size"])
  sizes.changed.both <- sizes.changed.both.d_a + sizes.changed.both.d_b

  # calc synchronicity degree based on changed lines within artefacts
  sync.degree.loc <- sizes.changed.both/sizes.changed

  return(c(sync.degree.artefacts, sync.degree.loc))
}

# Calculate the synchronicity degree for the commits of two developers within a burst, using the geometric mean
# Param: commits.d_a - the commits of developer a within the considered burst
# Param: commits.d_b - the commits of developer b within the same considered burst
# Param: analyzeFiles - whether to analyze files or not (then artefacts proximity or feature are analyzed)
# Returns: synchronicity degree using geometric mean on the fraction of sync. changed lines of a developer over all changed lines of a devloper
calculate.weighted.synchronicity.degree <- function(commits.d_a, commits.d_b, analyzeFiles) {

  # check whether this analysis is file-based or not
  artefact <- get.artefact.column(analyzeFiles)

  # get changed artefacts of d_a resp. d_b
  changed.artefacts.d_a <- unique(commits.d_a[,artefact])
  changed.artefacts.d_b <- unique(commits.d_b[,artefact])

  # get changed artefacts both
  changed.artefacts.both <- intersect(changed.artefacts.d_a, changed.artefacts.d_b)

  # do not consider "Base_Feature" or "File_Level" as commonly changed artefacts
  changed.artefacts.both <- changed.artefacts.both[!changed.artefacts.both %in% c("Base_Feature", "File_Level")]

  # get changed lines of d_a resp. d_b
  sizes.changed.d_a <- sum(commits.d_a[,"size"])
  sizes.changed.d_b <- sum(commits.d_b[,"size"])

  # get changed lines of artefacts changed by both
  sizes.changed.both.d_a <- sum(commits.d_a[commits.d_a[,artefact] %in% changed.artefacts.both,"size"])
  sizes.changed.both.d_b <- sum(commits.d_b[commits.d_b[,artefact] %in% changed.artefacts.both,"size"])

  # calc synchronicity degree using the geometric mean
  sync.degree <- sqrt((sizes.changed.both.d_a/sizes.changed.d_a) * (sizes.changed.both.d_b/sizes.changed.d_b))

  return(sync.degree)
}


## / / / / / / / / / / / / / /
## Email Bursts
## / / / / / / / / / / / / / /

# Determines email bursts.
# Returns a list which first element is the data frame of author pairs
# and which second element is a list containing data frames of  bursts for each author pair
determine.email.bursts <- function(emails, simulated=FALSE, thread.based=FALSE,
                                   based.on.commit.messages=FALSE, c.burst.data=NULL) {

  if(simulated) {
    pairs.of.authors.data.frame <- emails$pairs.of.authors

    # the following two lines are just a hack to use the same function parameters as for non-simulated burst identification
    authors <- pairs.of.authors.data.frame # pairs of authors
    emails.per.author <- emails$simulations # emails per pair
  } else {

    authors <- unique(emails[,"author"])

    # count.authors <- count(emails, "author")
    # authors <- as.vector(count.authors[count.authors["freq"] > 100,1])

    pairs.of.authors.data.frame <- as.data.frame(t(combn(authors, 2, simplify=TRUE)))
    colnames(pairs.of.authors.data.frame) <- c("d_a", "d_b")

    emails.per.author <- mclapply(as.vector(authors), mc.cores=CORES, FUN=function(author){
      return(emails[emails["author"] == author,])
    })
  }

  # call garbage collection
  rm(emails)
  # gc()

  pairs.of.authors <- as.list(as.data.frame(t(pairs.of.authors.data.frame)))

  if(PARALLEL.JOBS > 1) {
    pairs.of.authors.data.frame <- split(pairs.of.authors.data.frame, cut(seq(1,length(pairs.of.authors)), PARALLEL.JOBS))[[JOB.NUMBER]]
    pairs.of.authors <- split(pairs.of.authors, cut(seq(1,length(pairs.of.authors)), PARALLEL.JOBS))[[JOB.NUMBER]]

    if(simulated) {
      authors <- pairs.of.authors.data.frame
      rownames(authors) <- seq(length=nrow(authors))
      emails.per.author <- split(emails.per.author, cut(seq(1, length(emails.per.author)), PARALLEL.JOBS))[[JOB.NUMBER]]
    }
    loginfo(paste("Developer pairs to analyze in this job: ", length(pairs.of.authors)))
  }

  # apply burst detection to each pair of developers
  list.of.e.bursts <- mclapply(pairs.of.authors, mc.cores=CORES, emails.per.author=emails.per.author, authors=authors, FUN=function(pair, emails.per.author, authors){

    if(simulated) {
      # get the emails from the individual simulated time-series of the current pair
      pair.no <- get.pair.no(pair[1], pair[2], authors) # notice: 'authors' represents the pairs of authors
      emails.d_a <- as.data.frame(as.matrix(get.emails.of.author(pair[1], emails.per.author[[pair.no]]))) # notice: 'emails.per.author' contains all emails per pair!
      emails.d_b <- as.data.frame(as.matrix(get.emails.of.author(pair[2], emails.per.author[[pair.no]]))) # notice: 'emails.per.author' contains all emails per pair!
    } else {
      emails.d_a <- get.emails.by.author(pair[1], emails.per.author, authors) #emails[emails["author"] == pair[1],]
      emails.d_b <- get.emails.by.author(pair[2], emails.per.author, authors) #emails[emails["author"] == pair[2],]
    }

    # call garbage collection
    # rm(emails.per.author)
    # gc()

    email.bursts.a.b <- list()

    email.data.a.b <- list()

    if(based.on.commit.messages) {
          # get the c-burst data of the current pair
          pair.no.in.cburst.data <- get.pair.no(pair[1], pair[2], c.burst.data$pairs.of.developers.data.frame.from.c.bursts)
          if(length(pair.no.in.cburst.data) != 1) {
            pair.no.in.cburst.data <- get.pair.no(pair[2], pair[1], c.burst.data$pairs.of.developers.data.frame.from.c.bursts)
          }

          if(length(pair.no.in.cburst.data) < 1) {
            commit.bursts <- list()
          } else {
            commit.bursts <- c.burst.data$c.bursts[[pair.no.in.cburst.data]]
          }

          # iterate over all c-bursts of the current developer pair
          for(cb in commit.bursts) {
             # get beginning of commit message of each commit
             current.messages <- cb$messages$commitMessage
             current.messages <- tolower(trimws(current.messages))

             # get emails of d_a which email subject is similar to one of the commit messages
             emails.d_a.filtered <- emails.d_a[unlist(lapply(emails.d_a[["subject"]], FUN=function(email) {
                                                             reduced.subject <- sub("((((re)|(aw)): *)|(\\[[[:print:]]*\\] *)*)*", "",
                                                                                    tolower(trimws(email)))
                                                             if(nchar(reduced.subject) == 0) return(FALSE)
                                                             return(any(startsWith(current.messages, reduced.subject)))
						             })),]

             # get emails of d_b which email subject is similar to one of the commit messages
             emails.d_b.filtered <- emails.d_b[unlist(lapply(emails.d_b[["subject"]], FUN=function(email) {
                                                             reduced.subject <- sub("((((re)|(aw)): *)|(\\[[[:print:]]*\\] *)*)*", "",
                                                                                    tolower(trimws(email)))
                                                             if(nchar(reduced.subject) == 0) return(FALSE)
                                                             return(any(startsWith(current.messages, reduced.subject)))
	                                                     })),]

             ## the following statements would randomize the emails per c-burst, which does not make sense and does not work for smoothing
             #if(based.on.commit.messages && simulated) {
             #  emails.filtered.both <- rbind(emails.d_a.filtered, emails.d_b.filtered)
             #  emails.sim <- randomize.email.timestamps(pair, emails.filtered.both)
             #  emails.d_a.filtered <- get.emails.of.author(pair[1], emails.sim)
             #  emails.d_b.filtered <- get.emails.of.author(pair[2], emails.sim)
             #}

             emails <- list()
             emails$d_a <- emails.d_a.filtered
             emails$d_b <- emails.d_b.filtered
             email.data.a.b <- c(email.data.a.b, list(emails))
          }

    } else {
          emails <- list()
          emails$d_a <- emails.d_a
          emails$d_b <- emails.d_b
          email.data.a.b <- c(email.data.a.b, list(emails))
    }

    for(element in email.data.a.b) {

      emails.d_a <- element$d_a
      emails.d_b <- element$d_b

    if(nrow(emails.d_a) > 0 && nrow(emails.d_b) > 0) {

    # filter only the range of years in which both developers write emails (+/- 1 year)
    emails.d_a <- emails.d_a[year(emails.d_a[["creationDate"]]) %in%
                               (c((as.numeric(min(year(emails.d_b[["creationDate"]])))-1):(as.numeric(max(year(emails.d_b[["creationDate"]])))+1))),]

    if(thread.based) {
      emails.d_a.both.send.to.same.thread <- emails.d_a[emails.d_a[,"threadId"] %in% unique(emails.d_b[,"threadId"]),]
      if(nrow(emails.d_a.both.send.to.same.thread) == 0) {
		  emails.d_a = data.frame()
      }
    }

    if(nrow(emails.d_a) > 0) {
      # emails.d_b <- emails.d_b[year(emails.d_b[["creationDate"]]) %in%
      #                          c((as.numeric(min(year(emails.d_a[["creationDate"]])))-1):(as.numeric(max(year(emails.d_a[["creationDate"]])))+1)),]
      #
      # if(nrow(emails.d_b) > 0) {

      emails.d_a <- emails.d_a[ymd(as.Date(emails.d_a[["creationDate"]])) > ymd(as.Date(min(emails.d_b[["creationDate"]])) - days(EPSILON+1))
                               & ymd(as.Date(emails.d_a[["creationDate"]])) < ymd(as.Date(max(emails.d_b[["creationDate"]])) + days(EPSILON+1)),]

      if(nrow(emails.d_a) > 0) {

        emails.d_a <- split(emails.d_a, seq(1,nrow(emails.d_a)))

        lapply(emails.d_a, emails.d_b=emails.d_b, epsilon=EPSILON, FUN=function(e.d_a, emails.d_b, epsilon) {

		  if(thread.based) {
			  thread.id.d_a <- e.d_a[["threadId"]]

			  # select the emails of d_b in the specified time interval that belong to the same thread as the email of d_a
                          emails.d_b.same.thread <- emails.d_b[emails.d_b[,"threadId"]==thread.id.d_a,]

                          if(nrow(emails.d_b.same.thread) > 0) {
				# only consider emails of the same thread
				emails.d_b <- emails.d_b.same.thread
			  } else {
				# since there are no emails belonging to the same thread, remove all rows from emails.d_b
				emails.d_b <- emails.d_b[NULL,]
			  }
		  }

          emails.d_a[["creationDate"]] <- sapply(emails.d_a[["creationDate"]], FUN=function(x) fix.midnight(x), USE.NAMES=FALSE)
          emails.d_b[["creationDate"]] <- sapply(emails.d_b[["creationDate"]], FUN=function(x) fix.midnight(x), USE.NAMES=FALSE)

          creation.date.d_a <- ymd_hms(e.d_a["creationDate"])

          emails.d_b <- emails.d_b[ymd(as.Date(emails.d_b[,"creationDate"])) > ymd(as.Date(creation.date.d_a) - days(epsilon+1))
                                   & ymd(as.Date(emails.d_b[,"creationDate"])) < ymd(as.Date(creation.date.d_a) + days(epsilon+1)),]

          if(nrow(emails.d_b) > 0) {
            # select emails of d_b which were created at creation.date.d_a +/- epsilon days
            emails.d_b.same.time <- emails.d_b[abs(difftime(creation.date.d_a, ymd_hms(emails.d_b[,"creationDate"]))) <= duration(epsilon, "day"),]

            # call garbage collection
            # rm(e.d_a)
            # rm(emails.d_b)
            # gc()

            # check if there actually is an email burst. Only if there is a burst, add the burst to the list of bursts
            if(nrow(emails.d_b.same.time) > 0) {

              # determine beginning and end of email burst
              new.email.burst.start <- min(creation.date.d_a, ymd_hms(emails.d_b.same.time[,"creationDate"]))
              new.email.burst.end <- max(creation.date.d_a, ymd_hms(emails.d_b.same.time[,"creationDate"]))
              new.email.burst <- data.frame(start=new.email.burst.start,
                                            end=new.email.burst.end,
                                            occurence=new.email.burst.start+(new.email.burst.end-new.email.burst.start)/2)

              new.email.burst$start <- fix.midnight(new.email.burst$start)
              new.email.burst$end <- fix.midnight(new.email.burst$end)
              new.email.burst$occurence <- fix.midnight(new.email.burst$occurence)

              # call garbage collection
              # rm(emails.d_b.same.time)
              # rm(creation.date.d_a)
              # gc()

              # first check if email burst overlaps with some other email burst
              l <- length(email.bursts.a.b)
              bursts.to.remove <- c()

              if(l > 0) {
                for(i in c(1:length(email.bursts.a.b))) {
                  burst <- email.bursts.a.b[i][[1]]

                  # only check for overlaps if some preconditions are fulfilled
                  if(year(burst$end) >= year(new.email.burst$start) & year(burst$start) <= year(new.email.burst.end)) {

                    burst.start <-ymd_hms(burst$start)
                    burst.end <-ymd_hms(burst$end)
                    new.email.burst.start <-ymd_hms(new.email.burst$start)
                    new.email.burst.end <-ymd_hms(new.email.burst$end)

                    if(int_overlaps(
                      interval(burst.start, burst.end),
                      interval(new.email.burst.start, new.email.burst.end)
                    )) {

                      new.email.burst.start <- min(burst.start, new.email.burst.start)
                      new.email.burst.end <- max(burst.end, new.email.burst.end)
                      new.email.burst <- data.frame(start=new.email.burst.start,
                                                    end=new.email.burst.end,
                                                    occurence=new.email.burst.start+(new.email.burst.end-new.email.burst.start)/2)

                      new.email.burst$start <- fix.midnight(new.email.burst$start)
                      new.email.burst$end <- fix.midnight(new.email.burst$end)
                      new.email.burst$occurence <- fix.midnight(new.email.burst$occurence)

                      bursts.to.remove <- c(bursts.to.remove, i)
                    }
                  }
                }
              }

              # add new email burst (or merged one) to list of bursts for the current pair of authors
              if(length(bursts.to.remove) > 0) {
                email.bursts.a.b <<- c(email.bursts.a.b[-bursts.to.remove], list(new.email.burst))
              } else {
                email.bursts.a.b <<- c(email.bursts.a.b, list(new.email.burst))
              }

              # rm(bursts.to.remove)
              # rm(new.email.burst)
              # gc()
            }
          } else {
            # rm(emails.d_b)
            # rm(e.d_a)
            # gc()
          }

          # gc()
        })

      } else {
        # rm(emails.d_a)
        # rm(emails.d_b)
        # gc()
      }
      # } else {
      #   rm(emails.d_a)
      #   rm(emails.d_b)
      #   gc()
      #  }
    } else {
      # rm(emails.d_a)
      # rm(emails.d_b)
      # gc()
    }
    }
    } ## end of for-loop over "elements"
    # combine all list elements to one data frame
    email.bursts.a.b <- rbind.fill(email.bursts.a.b)
    # print(email.bursts.a.b)
    loginfo(paste("finished developer pair: ", pair[1], pair[2]))
    return(email.bursts.a.b)
  })

  #  for(x in rownames(pairs.of.authors.data.frame)) {
  #    print(pairs.of.authors.data.frame[x,])
  #    print(map.of.e.bursts[[x]])
  # }

  # return [1] data frame containing the pairs of authors
  # return [2] list which contains a data frame of bursts for each pair of authors
  return(list(pairs.of.authors.data.frame,list.of.e.bursts))
}


# Analyzes number of bursts per pair, length of bursts
# Results are printed to console
analyze.email.bursts <- function(pairs.of.authors.data.frame, list.of.e.bursts, emails, output.path.dat, output.path.txt, simulated=FALSE) {

  if(!simulated) {
    authors <- unique(emails[,"author"])
  } else {
    authors <- unique(unlist(as.list(pairs.of.authors.data.frame)))
  }

  pairs.of.authors <- as.list(as.data.frame(t(pairs.of.authors.data.frame)))

  # get overall number of bursts
  number.bursts.overall <- get.number.of.bursts.overall(list.of.e.bursts)

  # iterate over all pairs to determine the number of e-bursts per pair
  number.of.intervals <- get.number.of.bursts.per.pair(pairs.of.authors, pairs.of.authors.data.frame, list.of.e.bursts)

  # iterate over all e-bursts for each pair to determine the lengths of bursts
  length.of.intervals <- get.lengths.of.bursts(pairs.of.authors, pairs.of.authors.data.frame, list.of.e.bursts)

  # iterate over all e-bursts to determine the number of communication activities per burst
  if(!simulated) {
    email.activities.per.burst <- get.activities.of.bursts(pairs.of.authors, pairs.of.authors.data.frame, list.of.e.bursts, "email", emails, "creationDate")
  }

  # Save calculated numbers to file
  data <- list()
  data$pairs <- pairs.of.authors.data.frame
  data$bursts.per.pair <- number.of.intervals
  data$length.of.bursts <- length.of.intervals
  if(!simulated) {
    data$activities.per.burst <- email.activities.per.burst
  }

  save(data, file=output.path.dat)

  # Generate output for all determined results (from all the lists)

  output.number.of.authors <- paste("Number of authors of emails:", length(authors), sep=" ")
  output.number.of.bursts.overall <- paste("Number of bursts (overall):", number.bursts.overall, sep=" ")

  output.median.number.of.bursts.per.pair <- paste("Number of e-bursts per pair: (median)", median(number.of.intervals), sep=" ")
  output.mean.number.of.bursts.per.pair <- paste("Number of e-bursts per pair: (mean)", mean(number.of.intervals), sep=" ")
  output.min.number.of.bursts.per.pair <- paste("Number of e-bursts per pair: (min)", min(number.of.intervals), sep=" ")
  output.max.number.of.bursts.per.pair <- paste("Number of e-bursts per pair: (max)", max(number.of.intervals), sep=" ")

  output.median.length.of.bursts <- paste("Length of e-bursts: (median)", median(length.of.intervals), sep=" ")
  output.mean.length.of.bursts <- paste("Length of e-bursts: (mean)", mean(length.of.intervals), sep=" ")
  output.min.length.of.bursts <- paste("Length of e-bursts: (min)", min(length.of.intervals), sep=" ")
  output.max.length.of.bursts <- paste("Length of e-bursts: (max)", max(length.of.intervals), sep=" ")

  if(!simulated) {
    output.median.email.activities.per.burst <- paste("Email activities per burst: (median)", median(email.activities.per.burst), sep=" ")
    output.mean.email.activities.per.burst <- paste("Email activities per burst: (mean)", mean(email.activities.per.burst), sep=" ")
    output.min.email.activities.per.burst <- paste("Email activities per burst: (min)", min(email.activities.per.burst), sep=" ")
    output.max.email.activities.per.burst <- paste("Email activities per burst: (max)", max(email.activities.per.burst), sep=" ")

    output <- paste(output.number.of.authors, output.number.of.bursts.overall,
                    output.median.number.of.bursts.per.pair, output.mean.number.of.bursts.per.pair, output.min.number.of.bursts.per.pair, output.max.number.of.bursts.per.pair,
                    output.median.email.activities.per.burst, output.mean.email.activities.per.burst, output.min.email.activities.per.burst, output.max.email.activities.per.burst,
                    output.median.length.of.bursts, output.mean.length.of.bursts, output.min.length.of.bursts, output.max.length.of.bursts,"",
                    sep="\n")
  } else {
    output <- paste(output.number.of.authors, output.number.of.bursts.overall,
                    output.median.number.of.bursts.per.pair, output.mean.number.of.bursts.per.pair, output.min.number.of.bursts.per.pair, output.max.number.of.bursts.per.pair,
                    output.median.length.of.bursts, output.mean.length.of.bursts, output.min.length.of.bursts, output.max.length.of.bursts,"",
                    sep="\n")
  }
  cat(output)
  write(output, file=output.path.txt)
}
