## Copyright (c) Thomas Bock, 2016-2017
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

# STAND ALONE APPLICATION

## / / / / / / / / / / / / / /
## Set environment for script
## / / / / / / / / / / / / / /

options(stringsAsFactors = FALSE)
options(warn = 1)

# determine path of this very script file
getScriptPath <- function(){
  cmd.args <- commandArgs()
  m <- regexpr("(?<=^--file=).+", cmd.args, perl=TRUE)
  script.dir <- dirname(regmatches(cmd.args, m))
  if(length(script.dir) == 0) stop("can't determine script dir: please call the script with Rscript")
  if(length(script.dir) > 1) stop("can't determine script dir: more than one '--file' argument detected")
  return(script.dir)
}

source(file.path(getScriptPath(), "util.R"), chdir = TRUE) # get utility functions
# source(file.path(getScriptPath(), "coordinationBursts.R"), chdir = TRUE) # get functions to identify co-commit and email bursts
# source(file.path(getScriptPath(), "analysis.R"), chdir = TRUE) # get functions to run several kinds of analyses
# source(file.path(getScriptPath(), "simulation.R"), chdir = TRUE) # get function to generate simulated time-series

## libraries
#suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
#suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(logging)) # logging
#suppressWarnings(library(matrixStats)) # colMedians
#suppressWarnings(library(lsr)) # cohen's d effect size
#suppressWarnings(library(effsize)) # cliff's delta effect size

basicConfig()
#addHandler(writeToConsole)

## / / / / / / / / / / / / / /
## Configuration parameters
## / / / / / / / / / / / / / /

## Read command line parameters
if (!exists("cargs"))
  cargs = commandArgs(trailingOnly = TRUE)

if(length(cargs)==5) {

  # case-study
  CASESTUDY <- cargs[2]

  # directory to read input and write output files
  PATH.TO.DATA <- paste0(cargs[1], CASESTUDY) #"~/MA/qemu"

  # directory to read R object data of simulations
  PATH.TO.SIMULATION.DATA <- file.path(PATH.TO.DATA, "simulation")

  # analysis type (files, features, functions)
  ANALYSIS <- cargs[3]

  # the maximal allowed time difference between two commits or two mails to be considered within a burst
  EPSILON  <- as.numeric(cargs[4])

  THREADS <- as.numeric(cargs[5])
  if(THREADS < 0 || THREADS > 1) stop("Invalid argument: Threads must be 0 or 1")

} else {
  stop(paste("Invalid number of arguments. Please run the script with the following args: \n",
             "% Rscript correlation.R path.to.data casestudy analysis epsilon threads \n", "Set the threads parameter to 1 if you want to analyze thread-based E-bursts, otherwise set this parameter to 0.", sep=" "))
}

loginfo(paste0("CASESTUDY=", CASESTUDY))
loginfo(paste0("EPSILON=", EPSILON))
loginfo(paste0("THREADS=", THREADS))
loginfo(paste0("ANALYSIS=", ANALYSIS))

COMMIT.MESSAGES = paste0("commitMessages_", ANALYSIS, "s_") # ""

correlation.test <- function(){
  # load c-burst data
  load(paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", ANALYSIS,"s_", EPSILON, ".dat"))
  developer.pairs <- data$pairs
  c.bursts.per.pair <- data$bursts.per.pair

  # load e-burst data
  if(THREADS==0) {
    load(paste0(PATH.TO.DATA, "/", "e_bursts_", COMMIT.MESSAGES , "analysis_", EPSILON, ".dat"))
  } else {
    load(paste0(PATH.TO.DATA, "/", "e_bursts_thread_", COMMIT.MESSAGES, "analysis_", EPSILON, ".dat"))
  }
  author.pairs <- data$pairs
  e.bursts.per.pair <- data$bursts.per.pair

  # iterate over developer pairs and check if authors' pair exists
  developer.email.bursts.per.pair <- vector(mode="logical", length=nrow(developer.pairs))
  for(i in c(1:nrow(developer.pairs))) {
    email.pair.no <- get.pair.no(developer.pairs[i,1], developer.pairs[i,2], author.pairs)

    if(length(email.pair.no) != 1) {
      email.pair.no <- get.pair.no(developer.pairs[i,2], developer.pairs[i,1], author.pairs)
    }
    if(length(email.pair.no) != 1) {
      developer.email.bursts.per.pair[i] <- 0
    } else {
      developer.email.bursts.per.pair[i] <- e.bursts.per.pair[[email.pair.no]]
    }
  }

  loginfo("Perform correlation test...")
  # calculate correlation test
  cor.test.result <- cor.test(c.bursts.per.pair, developer.email.bursts.per.pair)

  loginfo("Fit linear model...")
  # linear model
  model <- lm(c.bursts.per.pair ~ developer.email.bursts.per.pair)

  # Goodness of fit
  summary <- summary(model)

  # save number of corresponding c- and e-bursts in one R object file
  corr.data <- list()
  corr.data$c.bursts.per.pair <- c.bursts.per.pair
  corr.data$developer.email.bursts.per.pair <- developer.email.bursts.per.pair
  corr.data$lm <- model
  corr.data$summary <- summary

  if(THREADS==0) {
    output.path <- paste0(PATH.TO.DATA, "/", "correlation_analysis_", ANALYSIS,"s_", EPSILON, ".dat")
  } else {
    output.path <- paste0(PATH.TO.DATA, "/", "correlation_thread_analysis_", ANALYSIS,"s_", EPSILON, ".dat")
  }
  loginfo(paste("Saved correlated data and linear model to:", output.path))
  save(corr.data, file=output.path)

  return(list(cor.test.result,model,summary))
}

result <- correlation.test()

print(result)
