## Copyright (c) Thomas Bock, 2016-2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## / / / / / / / / / / / / / /
## Set environment for script
## / / / / / / / / / / / / / /

options(stringsAsFactors = FALSE)
options(warn = 1)

# determine path of this very script file
getScriptPath <- function(){
    cmd.args <- commandArgs()
    m <- regexpr("(?<=^--file=).+", cmd.args, perl=TRUE)
    script.dir <- dirname(regmatches(cmd.args, m))
    if(length(script.dir) == 0) stop("can't determine script dir: please call the script with Rscript")
    if(length(script.dir) > 1) stop("can't determine script dir: more than one '--file' argument detected")
    return(script.dir)
}

source(file.path(getScriptPath(), "util.R"), chdir = TRUE) # get utility functions
source(file.path(getScriptPath(), "coordinationBursts.R"), chdir = TRUE) # get functions to identify co-commit and email bursts
source(file.path(getScriptPath(), "analysis.R"), chdir = TRUE) # get functions to run several kinds of analyses
source(file.path(getScriptPath(), "codegrowth.R"), chdir = TRUE) # get functions to determine synchronous and non-synchronous commits and calculate code growth metrics
source(file.path(getScriptPath(), "simulation.R"), chdir = TRUE) # get functions to generate simulated time-series
source(file.path(getScriptPath(), "responseTimes.R"), chdir = TRUE) # get functions to compute response times
source(file.path(getScriptPath(), "qualitativeAnalysis.R"), chdir = TRUE) # get functions to prepare qualitative analysis
source(file.path(getScriptPath(), "qualitativeAnalysisEvaluation.R"), chdir = TRUE) # get functions to evaluate qualitative analysis
source(file.path(getScriptPath(), "qualitativeAnalysisAutomaticComparison.R"), chdir = TRUE) # get functions to perform automatic comparison of commit messages and e-mail subjects
#source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(logging)) # logging

basicConfig()
#addHandler(writeToConsole)

## / / / / / / / / / / / / / /
## Configuration parameters
## / / / / / / / / / / / / / /

## Read command line parameters
if (!exists("cargs"))
  cargs = commandArgs(trailingOnly = TRUE)

if(length(cargs)==7) {

  # case-study
  CASESTUDY <- cargs[2]

  # directory to read input and write output files
  PATH.TO.DATA <- paste0(cargs[1], CASESTUDY) #"~/MA/qemu"

  # directory to save and read R object data of analysis parts
  PATH.TO.PART.DATA <- file.path(PATH.TO.DATA, "part.data")

  # analysis type (file, feature, function, email)
  ANALYSIS <- cargs[3]

  # the maximal allowed time difference between two commits or two mails to be considered within a burst
  EPSILON  <- as.numeric(cargs[4])

  # number of jobs, i.e. in how much sub-jobs the analysis should be divided
  PARALLEL.JOBS <- as.numeric(cargs[5])

  if(PARALLEL.JOBS > 1 && !file.exists(PATH.TO.PART.DATA) && ANALYSIS != "commit-simulation" && ANALYSIS != "email-simulation") {
    dir.create(PATH.TO.PART.DATA)
  }

  # if there are more than one jobs, the job number determines which part of the whole analysis should be executed
  JOB.NUMBER = cargs[6]

  # check wheter JOB.NUMBER is not "final". Otherwise there is no actual job number.
  if(JOB.NUMBER != "final") {

    # check whether JOB.NUMBER allready is a number or whether it must be read from environment variable
    if(suppressWarnings(!is.na(as.numeric(JOB.NUMBER)))) {
      # JOB.NUMBER allready is a number
      JOB.NUMBER = as.numeric(JOB.NUMBER)
    } else {
      # JOB.NUMBER must be read from environment variable
      JOB.NUMBER = as.numeric(Sys.getenv(JOB.NUMBER))
    }

    if(JOB.NUMBER > PARALLEL.JOBS) {
      stop("Job number exceeds number of parallel jobs!")
    }
  }

  # number of cores (hyperthreads) for parallel execution
  CORES <- cargs[7]
  if(CORES == "max") {
    CORES = max(4L, detectCores(), na.rm = TRUE) # use the maximal available number of cores
  } else if(CORES=="maxram") { # use only 40% of the maximal availble number of cores to prevent exceeding the memory limit
    CORES = max(4L, detectCores(), na.rm = TRUE)
    CORES = round(0.4 * CORES)
  } else {
    CORES = as.numeric(CORES)
  }

} else {
  stop(paste("Invalid number of arguments. Please run the script with the following args: \n",
             "% Rscript main.R path.to.data casestudy analysis epsilon jobs jobnumer cores", sep=" "))
}

## Set other configuration paramters

# specify end date of the whole analysis (or NULL to take the latest possible date)
END.DATE <- ymd_hms("2016-07-27 23:59:59")

# specify start date of the whole analyis (or NULL to take the earliest possible date)
START.DATE <- ymd_hms("1990-01-01 00:00:01") #NULL

BASED.ON.COMMIT.MESSAGES <- TRUE

## / / / / / / / / / / / / / /
## Get input data
## / / / / / / / / / / / / / /

path.to.proximity.commit.list <- file.path(PATH.TO.DATA, "c_proximity.list")
path.to.feature.commit.list   <- file.path(PATH.TO.DATA, "c_feature.list")
path.to.email.list            <- file.path(PATH.TO.DATA, "e_proximity.list")

commits.proximity <- read.commit.data(path.to.proximity.commit.list, "proximity")
commits.feature <- read.commit.data(path.to.feature.commit.list, "feature")

if(ANALYSIS=="email" || ANALYSIS=="email-simulation" || ANALYSIS == "email-thread-based"
   || startsWith(ANALYSIS, "email-commitMessages-")
   || ANALYSIS=="smooth-file-email"
   || ANALYSIS=="smooth-feature-email" || ANALYSIS=="smooth-function-email"
   || ANALYSIS=="smooth-file-email-threads"
   || ANALYSIS=="smooth-feature-email-threads" || ANALYSIS=="smooth-function-email-threads"
   || ANALYSIS=="response-times-email-threads" || ANALYSIS =="response-times-email-global"
   || startsWith(ANALYSIS,"qual-")) {

  emails <- read.email.data(path.to.email.list)

  # disregard emails which authors are not authors of a commit
  emails <- filter.commit.authors(emails, commits.proximity)
}

# disregard commits and emails before START.DATE and after END.DATE
commits.proximity <- filter.data.by.date.interval(commits.proximity, "authorDate", END.DATE, START.DATE)
commits.feature <- filter.data.by.date.interval(commits.feature, "authorDate", END.DATE, START.DATE)
if(ANALYSIS=="email" || ANALYSIS=="email-simulation" || ANALYSIS == "email-thread-based"
   || startsWith(ANALYSIS, "email-commitMessages-")
   || ANALYSIS=="smooth-file-email"
   || ANALYSIS=="smooth-feature-email" || ANALYSIS=="smooth-function-email"
   || ANALYSIS=="smooth-file-email-threads"
   || ANALYSIS=="smooth-feature-email-threads" || ANALYSIS=="smooth-function-email-threads"
   || ANALYSIS=="response-times-email-threads" || ANALYSIS =="response-times-email-global"
   || startsWith(ANALYSIS,"qual-")) emails <- filter.data.by.date.interval(emails, "creationDate", END.DATE, START.DATE)

## / / / / / / / / / / / / / /
## Identify bursts
## / / / / / / / / / / / / / /

loginfo(paste0("CASESTUDY=", CASESTUDY))
loginfo(paste0("EPSILON=", EPSILON))
loginfo(paste0("CORES=", CORES))
loginfo(paste0("JOBS=", PARALLEL.JOBS))
loginfo(paste0("JOB.NUMBER=", JOB.NUMBER))

if(ANALYSIS == "file") {
  # File-based co-commit bursts
  loginfo("START: C-BURSTS FILES")
  do.analysis.c.bursts("file", "files", commits.proximity, analyze.files = TRUE)

} else if(ANALYSIS == "feature") {
  # Feature-based co-commit bursts
  loginfo("START: C-BURSTS FEATURES")
  do.analysis.c.bursts("feature", "features", commits.feature, analyze.files = FALSE)

} else if(ANALYSIS == "function") {
  # Function-based co-commit bursts
  loginfo("START: C-BURSTS FUNCTIONS")
  do.analysis.c.bursts("function", "functions", commits.proximity, analyze.files = FALSE)

} else if(ANALYSIS == "email") {
  # Email bursts
  loginfo("START: E-BURSTS")
  do.analysis.e.bursts(emails)

} else if(ANALYSIS == "email-thread-based") {
  # Email bursts
  loginfo("START: E-BURSTS THREAD-BASED")
  do.analysis.e.bursts(emails, thread.based=TRUE)

} else if(ANALYSIS == "email-simulation") {
  # Generate simulated email activities
  loginfo("START: GENERATE EMAIL SIMULATION")
  do.email.randomization(emails, "e_simulation.dat")

} else if(ANALYSIS == "email-sim") {
  # Email bursts
  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES)")
  do.analysis.e.bursts(NULL, simulated=TRUE)

} else if(ANALYSIS == "email-sim-thread-based") {
  # Email bursts
  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES)")
  do.analysis.e.bursts(NULL, simulated=TRUE, thread.based=TRUE)

} else if(ANALYSIS == "commit-simulation") {
  # Generate simulated commit activities
  loginfo("START: GENERATE COMMIT SIMULATION")
  do.commit.randomization(commits.proximity, "c_proximity.list", commits.feature, "c_feature.list")

} else if(ANALYSIS == "code-growth-file") {
  # Identify synchronous and non-synchronous commits based on file collaboration and calculate code growth and effort
  loginfo("START: CALCULATE CODE-GROWTH METRICS (FILE-BASED)")
  do.code.growth.analysis("file", "files", commits.proximity, analyze.files = TRUE)

} else if(ANALYSIS == "code-growth-feature") {
  # Identify synchronous and non-synchronous commits based on feature collaboration and calculate code growth and effort
  loginfo("START: CALCULATE CODE-GROWTH METRICS (FEATURE-BASED")
  do.code.growth.analysis("feature", "features", commits.feature, analyze.files = FALSE)

} else if(ANALYSIS == "code-growth-function") {
  # Identify synchronous and non-synchronous commits based on function collaboration and calculate code growth and effort
  loginfo("START: CALCULATE CODE-GROWTH METRICS (FUNCTION-BASED)")
  do.code.growth.analysis("function", "functions", commits.proximity, analyze.files = FALSE)

} else if(ANALYSIS == "smooth-file-email") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of file-based bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FILE-BASED)")
  do.smoothing("file", "files", commits.proximity, analyze.files = TRUE, "email", simulated=FALSE, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, emails)

} else if(ANALYSIS == "smooth-feature-email") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of feature-based bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FEATURE-BASED")
  do.smoothing("feature", "features", commits.feature, analyze.files = FALSE, "email", simulated=FALSE, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, emails)

} else if(ANALYSIS == "smooth-function-email") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of function-based bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FUNCTION-BASED)")
  do.smoothing("function", "functions", commits.proximity, analyze.files = FALSE, "email", simulated=FALSE, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, emails)

} else if(ANALYSIS == "smooth-file-email-sim") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of email bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FILE-BASED, EMAIL-SIMULATION)")
  do.smoothing("file", "files", commits.proximity, analyze.files = TRUE, "email-sim", simulated=TRUE, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, NULL)

} else if(ANALYSIS == "smooth-feature-email-sim") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of simulated email bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FEATURE-BASED, EMAIL-SIMULATION)")
  do.smoothing("feature", "features", commits.feature, analyze.files = FALSE, "email-sim", simulated=TRUE, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, NULL)

} else if(ANALYSIS == "smooth-function-email-sim") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of simulated email bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FUNCTION-BASED, EMAIL-SIMULATION)")
  do.smoothing("function", "functions", commits.proximity, analyze.files = FALSE, "email-sim", simulated=TRUE, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, NULL)

} else if(ANALYSIS == "smooth-file-email-threads") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of file-based bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FILE-BASED, THREAD-BASED)")
  do.smoothing("file", "files", commits.proximity, analyze.files = TRUE, "email", simulated=FALSE, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, emails)

} else if(ANALYSIS == "smooth-feature-email-threads") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of feature-based bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FEATURE-BASED, THREAD-BASED")
  do.smoothing("feature", "features", commits.feature, analyze.files = FALSE, "email", simulated=FALSE, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, emails)

} else if(ANALYSIS == "smooth-function-email-threads") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of function-based bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FUNCTION-BASED, THREAD-BASED)")
  do.smoothing("function", "functions", commits.proximity, analyze.files = FALSE, "email", simulated=FALSE, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, emails)

} else if(ANALYSIS == "smooth-file-email-sim-threads") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of email bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FILE-BASED, THREAD-BASED, EMAIL-SIMULATION)")
  do.smoothing("file", "files", commits.proximity, analyze.files = TRUE, "email-sim", simulated=TRUE, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, NULL)

} else if(ANALYSIS == "smooth-feature-email-sim-threads") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of simulated email bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FEATURE-BASED, THREAD-BASED, EMAIL-SIMULATION)")
  do.smoothing("feature", "features", commits.feature, analyze.files = FALSE, "email-sim", simulated=TRUE, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, NULL)

} else if(ANALYSIS == "smooth-function-email-sim-threads") {
  source(file.path(getScriptPath(), "smoothedCurves.R"), chdir = TRUE) # get functions to generate smoothed curves
  # Generate smoothed curves of simulated email bursts
  loginfo("START: GENERATE SMOOTHED CURVES (FUNCTION-BASED, THREAD-BASED, EMAIL-SIMULATION)")
  do.smoothing("function", "functions", commits.proximity, analyze.files = FALSE, "email-sim", simulated=TRUE, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES, NULL)

} else if(ANALYSIS == "response-times-email-global") {
  # Compute response times in emails (global)
  loginfo("COMPUTE RESPONSE TIMES (EMAIL-GLOBAL)")
  compute.response.times("emails", "global", emails, NULL)

} else if(ANALYSIS == "response-times-email-threads") {
  # Compute response times in email threads
  loginfo("COMPUTE RESPONSE TIMES (EMAIL-THREADS)")
  compute.response.times("emails", "threads", emails, NULL)

} else if(ANALYSIS == "response-times-commits-files") {
  # Compute response times in commits on same files
  loginfo("COMPUTE RESPONSE TIMES (COMMITS, FILE-BASED)")
  compute.response.times("commits", "files", commits.proximity, analyze.files = TRUE)

} else if(ANALYSIS == "response-times-commits-features") {
  # Compute response times in commits on same features
  loginfo("COMPUTE RESPONSE TIMES (COMMITS, FEATURE-BASED)")
  compute.response.times("commits", "features", commits.feature, analyze.files = FALSE)

} else if(ANALYSIS == "response-times-commits-functions") {
  # Compute response times in commits on same functions
  loginfo("COMPUTE RESPONSE TIMES (COMMITS, FUNCTION-BASED)")
  compute.response.times("commits", "functions", commits.proximity, analyze.files = FALSE)

} else if(ANALYSIS == "qual-file") {
  # Prepare qualitative analysis on files and mails
  loginfo("PREPARE QUALITATIVE ANALYSIS (FILES, MAILS)")
  do.qualitative.analysis.c.e.bursts("file", "files", commits.proximity, analyze.files = TRUE, emails, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES)

} else if(ANALYSIS == "qual-file-thread") {
  # Prepare qualitative analysis on files and thread-based mails
  loginfo("PREPARE QUALITATIVE ANALYSIS (FILES, THREADS)")
  do.qualitative.analysis.c.e.bursts("file", "files", commits.proximity, analyze.files = TRUE, emails, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES)

} else if(ANALYSIS == "qual-feature") {
  # Prepare qualitative analysis on features and mails
  loginfo("PREPARE QUALITATIVE ANALYSIS (FEATURES, MAILS)")
  do.qualitative.analysis.c.e.bursts("feature", "features", commits.feature, analyze.files = FALSE, emails, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES)

} else if(ANALYSIS == "qual-feature-thread") {
  # Prepare qualitative analysis on features and thread-based mails
  loginfo("PREPARE QUALITATIVE ANALYSIS (FEATURES, THREADS)")
  do.qualitative.analysis.c.e.bursts("feature", "features", commits.feature, analyze.files = FALSE, emails, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES)

} else if(ANALYSIS == "qual-function") {
  # Prepare qualitative analysis on functions and mails
  loginfo("PREPARE QUALITATIVE ANALYSIS (FUNCTIONS, MAILS)")
  do.qualitative.analysis.c.e.bursts("function", "functions", commits.proximity, analyze.files = FALSE, emails, thread.based=FALSE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES)

} else if(ANALYSIS == "qual-function-thread") {
  # Prepare qualitative analysis on functions and thread-based mails
  loginfo("PREPARE QUALITATIVE ANALYSIS (FUNCTIONS, THREADS)")
  do.qualitative.analysis.c.e.bursts("function", "functions", commits.proximity, analyze.files = FALSE, emails, thread.based=TRUE, based.on.commit.messages=BASED.ON.COMMIT.MESSAGES)

} else if(ANALYSIS == "qual-evaluation-file") {
  # Evaluate qualitative analysis on files and mails
  loginfo("EVALUATE QUALITATIVE ANALYSIS (FILES, MAILS)")
  do.qualitative.analysis.evaluation("files", thread.based=FALSE)

} else if(ANALYSIS == "qual-evaluation-file-thread") {
  # Evaluate qualitative analysis on files and thread-based mails
  loginfo("EVALUATE QUALITATIVE ANALYSIS (FILES, THREADS)")
  do.qualitative.analysis.evaluation("files", thread.based=TRUE)

} else if(ANALYSIS == "qual-evaluation-feature") {
  # Evaluate qualitative analysis on features and mails
  loginfo("EVALUATE QUALITATIVE ANALYSIS (FEATURES, MAILS)")
  do.qualitative.analysis.evaluation("features", thread.based=FALSE)

} else if(ANALYSIS == "qual-evaluation-feature-thread") {
  # Evaluate qualitative analysis on features and thread-based mails
  loginfo("EVALUATE QUALITATIVE ANALYSIS (FEATURES, THREADS)")
  do.qualitative.analysis.evaluation("features", thread.based=TRUE)

} else if(ANALYSIS == "qual-evaluation-function") {
  # Evaluate qualitative analysis on functions and mails
  loginfo("EVALUATE QUALITATIVE ANALYSIS (FUNCTIONS, MAILS)")
  do.qualitative.analysis.evaluation("functions", thread.based=FALSE)

} else if(ANALYSIS == "qual-evaluation-function-thread") {
  # Evaluate qualitative analysis on functions and thread-based mails
  loginfo("EVALUATE QUALITATIVE ANALYSIS (FUNCTIONS, THREADS)")
  do.qualitative.analysis.evaluation("functions", thread.based=TRUE)

} else if(ANALYSIS == "qual-automatic-comparison-file") {
  # Compare e-mail subjects and commit messages on files and mails
  loginfo("COMPARE E-MAIL SUBJECTS AND COMMITT MESSAGES (FILES, MAILS)")
  do.qualitative.analysis.automatic.comparison("files", thread.based=FALSE)

} else if(ANALYSIS == "qual-automatic-comparison-file-thread") {
  # Compare e-mail subjects and commit messages on files and thread-based mails
  loginfo("COMPARE E-MAIL SUBJECTS AND COMMITT MESSAGES (FILES, THREADS)")
  do.qualitative.analysis.automatic.comparison("files", thread.based=TRUE)

} else if(ANALYSIS == "qual-automatic-comparison-feature") {
  # Compare e-mail subjects and commit messages on features and mails
  loginfo("COMPARE E-MAIL SUBJECTS AND COMMITT MESSAGES (FEATURES, MAILS)")
  do.qualitative.analysis.automatic.comparison("features", thread.based=FALSE)

} else if(ANALYSIS == "qual-automatic-comparison-feature-thread") {
  # Compare e-mail subjects and commit messages on features and thread-based mails
  loginfo("COMPARE E-MAIL SUBJECTS AND COMMITT MESSAGES (FEATURES, THREADS)")
  do.qualitative.analysis.automatic.comparison("features", thread.based=TRUE)

} else if(ANALYSIS == "qual-automatic-comparison-function") {
  # Compare e-mail subjects and commit messages on functions and mails
  loginfo("COMPARE E-MAIL SUBJECTS AND COMMITT MESSAGES (FUNCTIONS, MAILS)")
  do.qualitative.analysis.automatic.comparison("functions", thread.based=FALSE)

} else if(ANALYSIS == "qual-automatic-comparison-function-thread") {
  # Compare e-mail subjects and commit messages on functions and thread-based mails
  loginfo("COMPARE E-MAIL SUBJECTS AND COMMITT MESSAGES (FUNCTIONS, THREADS)")
  do.qualitative.analysis.automatic.comparison("functions", thread.based=TRUE)

} else if(ANALYSIS == "only-c-qual-file") {
	  # Prepare qualitative analysis on files
	  loginfo("PREPARE QUALITATIVE ANALYSIS (FILES)")
  do.qualitative.analysis.c.e.bursts("file", "files", commits.proximity, analyze.files = TRUE, emails, thread.based=FALSE, only.c.analysis=TRUE)

} else if(ANALYSIS == "only-c-qual-feature") {
	  # Prepare qualitative analysis on features
	  loginfo("PREPARE QUALITATIVE ANALYSIS (FEATURES)")
  do.qualitative.analysis.c.e.bursts("feature", "features", commits.feature, analyze.files = FALSE, emails, thread.based=FALSE, only.c.analysis=TRUE)

} else if(ANALYSIS == "only-c-qual-function") {
	  # Prepare qualitative analysis on functions
	  loginfo("PREPARE QUALITATIVE ANALYSIS (FUNCTIONS)")
  do.qualitative.analysis.c.e.bursts("function", "functions", commits.proximity, analyze.files = FALSE, emails, thread.based=FALSE, only.c.analysis=TRUE)

} else if(ANALYSIS == "email-commitMessages-file") {
  # Email bursts
  loginfo("START: E-BURSTS BASED ON COMMIT-MESSAGES OF FILE-BASED C-BURSTS")
  do.analysis.e.bursts(emails, based.on.commit.messages=TRUE, c.burst.artefact="files")

} else if(ANALYSIS == "email-commitMessages-file-thread-based") {
  # Email bursts
  loginfo("START: E-BURSTS THREAD-BASED BASED ON COMMIT-MESSAGES OF FILE-BASED C-BURSTS")
  do.analysis.e.bursts(emails, thread.based=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="files")

} else if(ANALYSIS == "email-commitMessages-feature") {
  # Email bursts
  loginfo("START: E-BURSTS BASED ON COMMIT-MESSAGES OF FEATURE-BASED C-BURSTS")
  do.analysis.e.bursts(emails, based.on.commit.messages=TRUE, c.burst.artefact="features")

} else if(ANALYSIS == "email-commitMessages-feature-thread-based") {
  # Email bursts
  loginfo("START: E-BURSTS THREAD-BASED BASED ON COMMIT-MESSAGES OF FEATURE-BASED C-BURSTS")
  do.analysis.e.bursts(emails, thread.based=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="features")

} else if(ANALYSIS == "email-commitMessages-function") {
  # Email bursts
  loginfo("START: E-BURSTS BASED ON COMMIT-MESSAGES OF FUNCTION-BASED C-BURSTS")
  do.analysis.e.bursts(emails, based.on.commit.messages=TRUE, c.burst.artefact="functions")

} else if(ANALYSIS == "email-commitMessages-function-thread-based") {
  # Email bursts
  loginfo("START: E-BURSTS THREAD-BASED BASED ON COMMIT-MESSAGES OF FUNCTION-BASED C-BURSTS")
  do.analysis.e.bursts(emails, thread.based=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="functions")

} else if(ANALYSIS == "email-sim-commitMessages-file") {
	  # Email bursts
	  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES) BASED ON COMMIT-MESSAGES OF FILE-BASED C-BURSTS")
  do.analysis.e.bursts(NULL, simulated=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="files")

} else if(ANALYSIS == "email-sim-commitMessages-file-thread-based") {
	  # Email bursts
	  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES) THREAD-BASED BASED ON COMMIT-MESSAGES OF FILE-BASED C-BURSTS")
  do.analysis.e.bursts(NULL, simulated=TRUE, thread.based=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="files")

} else if(ANALYSIS == "email-sim-commitMessages-feature") {
	  # Email bursts
	  loginfo("START: E-BURSTS BASED (OF SIMULATED TIME-SERIES) ON COMMIT-MESSAGES OF FEATURE-BASED C-BURSTS")
  do.analysis.e.bursts(NULL, simulated=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="features")

} else if(ANALYSIS == "email-sim-commitMessages-feature-thread-based") {
	  # Email bursts
	  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES) THREAD-BASED BASED ON COMMIT-MESSAGES OF FEATURE-BASED C-BURSTS")
  do.analysis.e.bursts(NULL, simulated=TRUE, thread.based=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="features")

} else if(ANALYSIS == "email-sim-commitMessages-function") {
	  # Email bursts
	  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES) BASED ON COMMIT-MESSAGES OF FUNCTION-BASED C-BURSTS")
  do.analysis.e.bursts(NULL, simulated=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="functions")

} else if(ANALYSIS == "email-sim-commitMessages-function-thread-based") {
	  # Email bursts
	  loginfo("START: E-BURSTS (OF SIMULATED TIME-SERIES) THREAD-BASED BASED ON COMMIT-MESSAGES OF FUNCTION-BASED C-BURSTS")
  do.analysis.e.bursts(NULL, simulated=TRUE, thread.based=TRUE, based.on.commit.messages=TRUE, c.burst.artefact="functions")

}

