## Copyright (c) Thomas Bock, 2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(parallel)) # tools for parallel execution

get.burst.commits <- function(commits, commit.messages, synchronous.commits, commit.bursts, pairs.of.developers.data.frame, analyzeFiles = FALSE) {

  # check whether this analysis is file-based or not
  artefact.column <- get.artefact.column(analyzeFiles)

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  burst.commits.per.pair <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(developer.pair) {
    loginfo(paste("C-burst analysis for developer pair:", developer.pair[1], developer.pair[2]))
    pair.commits <- commits[(commits[["author"]]==developer.pair[1]) | (commits[["author"]]==developer.pair[2]),]
    sync.pair.commits <- pair.commits[pair.commits[["commitHash"]] %in% synchronous.commits,]

    pair.no <- get.pair.no(developer.pair[1], developer.pair[2], pairs.of.developers.data.frame)
    bursts.data.frame <- commit.bursts[[pair.no]]
    burst.commit.data <- list()

    # Check if there is no burst for the current developer pair
    if(length(bursts.data.frame)>0) {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(burst in bursts) {
        ival <- interval(burst[1], burst[2])

        # Get the commits of both developers within current interval
        commits.d_a <- sync.pair.commits[sync.pair.commits[,"author"] == developer.pair[1] & ymd_hms(sync.pair.commits[,"authorDate"]) %within% ival,]
        commits.d_b <- sync.pair.commits[sync.pair.commits[,"author"] == developer.pair[2] & ymd_hms(sync.pair.commits[,"authorDate"]) %within% ival,]

        # check for common artefacts
        artefacts.d_a <- unique(commits.d_a[[artefact.column]])
        artefacts.d_b <- unique(commits.d_b[[artefact.column]])
        common.artefacts <- intersect(artefacts.d_a, artefacts.d_b)

        # commits changing common artefacts
        burst.commits <- list()
        burst.commits$commits <- rbind(commits.d_a[commits.d_a[[artefact.column]] %in% common.artefacts,],
                                       commits.d_b[commits.d_b[[artefact.column]] %in% common.artefacts,])
        burst.commits$artefacts <- common.artefacts

        # get commit messages for the burst
        commit.hashes <- unique(burst.commits$commits$commitHash)
        burst.commit.messages <- commit.messages[commit.messages[["commitHash"]] %in% commit.hashes,]
        burst.commits$messages <- burst.commit.messages

        burst.commit.data <- c(burst.commit.data, list(burst.commits))
      }
    }

    return(burst.commit.data) # return burst commits together with common artifacts
  })

  return(burst.commits.per.pair)
}

get.burst.emails <- function(emails, email.bursts, pairs.of.developers.data.frame, mboxparsing.data=NULL, thread.based=FALSE) {

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  burst.emails.per.pair <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(developer.pair) {
    loginfo(paste("E-burst analysis for developer pair:", developer.pair[1], developer.pair[2]))
    pair.emails <- emails[(emails[["author"]]==developer.pair[1]) | (emails[["author"]]==developer.pair[2]),]

    pair.no <- get.pair.no(developer.pair[1], developer.pair[2], pairs.of.developers.data.frame)
    bursts.data.frame <- email.bursts[[pair.no]]
    burst.email.data <- list()

    # Check if there is no burst for the current developer pair
    if(length(bursts.data.frame)>0) {

      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(burst in bursts) {
        ival <- interval(burst[1], burst[2])

        # Get the emails of both developers within current interval
        emails.d_a <- pair.emails[pair.emails[,"author"] == developer.pair[1] & ymd_hms(pair.emails[,"creationDate"]) %within% ival,]
        emails.d_b <- pair.emails[pair.emails[,"author"] == developer.pair[2] & ymd_hms(pair.emails[,"creationDate"]) %within% ival,]

        common.threads <- NULL
        if(thread.based) {
          # check for common threads
          threads.d_a <- unique(emails.d_a[["threadId"]])
          threads.d_b <- unique(emails.d_b[["threadId"]])
          common.threads <- intersect(threads.d_a, threads.d_b)

          emails.d_a <- emails.d_a[emails.d_a[["threadId"]] %in% common.threads,]
          emails.d_b <- emails.d_b[emails.d_b[["threadId"]] %in% common.threads,]
        }

        emails.mboxparsing <- NULL
        if(!is.null(mboxparsing.data)) {
          # select emails which are mentioned together with an artefact

          #emails.mboxparsing.d_a <- emails.d_a[emails.d_a[["messageId"]] %in% mboxparsing.data[["messageId"]],]
          emails.mboxparsing.d_a <- merge(x=emails.d_a, y=mboxparsing.data, by="messageId")

          #emails.mboxparsing.d_b <- emails.d_a[emails.d_b[["messageId"]] %in% mboxparsing.data[["messageId"]],]
          emails.mboxparsing.d_b <- merge(x=emails.d_b, y=mboxparsing.data, by="messageId")

          emails.mboxparsing <- rbind(emails.mboxparsing.d_a, emails.mboxparsing.d_b)
        }

        # burst messages
        burst.emails <- list()
        burst.emails$emails <- rbind(emails.d_a, emails.d_b)
        if(thread.based) burst.emails$threads <- common.threads
        if(!is.null(mboxparsing.data)) burst.emails$mboxparsing <- emails.mboxparsing
        burst.email.data <- c(burst.email.data, list(burst.emails))
      }
    }

    return(burst.email.data) # return burst emails together with common threads
  })

  return(burst.emails.per.pair)
}

get.developer.pairs.having.c.and.e.bursts <- function(developer.pairs.c, burst.commits, developer.pairs.e, burst.emails) {

  list.of.developer.pairs.c.bursts <- list()
  list.of.developer.pairs.e.bursts <- list()
  list.of.developer.pairs.both.bursts <- list()

  for(i in seq_along(burst.commits)) {
    if(length(burst.commits[i][[1]]) > 0) {

      developer.pair <- list()
      developer.pair$c.number <- i
      developer.pair$names <- unlist(developer.pairs.c[i,])
      developer.pair$commit.bursts <- length(burst.commits[i][[1]])
      list.of.developer.pairs.c.bursts <- c(list.of.developer.pairs.c.bursts, list(developer.pair))
    }
  }


  for(i in seq_along(burst.emails)) {
    if(length(burst.emails[i][[1]]) > 0) {

      developer.pair <- list()
      developer.pair$e.number <- i
      developer.pair$names <- unlist(developer.pairs.e[i,])
      developer.pair$email.bursts <- length(burst.emails[i][[1]])
      list.of.developer.pairs.e.bursts <- c(list.of.developer.pairs.e.bursts, list(developer.pair))
    }
  }

  for(i in seq_along(burst.commits)) {
    if(length(burst.commits[i][[1]]) > 0) {

      developer.pair <- list()
      developer.pair$names <- unlist(developer.pairs.c[i,])
      developer.pair$c.number <- i

      e.number <- get.pair.no(developer.pair$names[1], developer.pair$names[2], developer.pairs.e)
      if(length(e.number) < 1) e.number <- get.pair.no(developer.pair$names[2], developer.pair$names[1], developer.pairs.e)
      if(length(e.number) > 0 && length(burst.emails[e.number][[1]]) > 0) {
        developer.pair$e.number <- e.number
        developer.pair$commit.bursts <- length(burst.commits[i][[1]])
        developer.pair$email.bursts <- length(burst.emails[e.number][[1]])
        list.of.developer.pairs.both.bursts <- c(list.of.developer.pairs.both.bursts, list(developer.pair))
      }
    }
  }

  ## get pairs having more than 5 bursts each:
  numbers.in.this.list <- which(unlist(lapply(list.of.developer.pairs.both.bursts, FUN=function(x){
                                               (x$email.bursts > 5) && (x$commit.bursts > 5);
  })))
  print(numbers.in.this.list)

  return(list(list.of.developer.pairs.c.bursts, list.of.developer.pairs.e.bursts, list.of.developer.pairs.both.bursts))
}

