## Copyright (c) Thomas Bock, 2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(parallel)) # tools for parallel execution

do.qualitative.analysis.automatic.comparison <- function(artefact, thread.based=TRUE) {

  only.pairs.having.more.than.five.c.e.bursts <- FALSE

  if(thread.based) {
    thread = "thread_"
  } else {
    thread = ""
  }
  file.name.devCEbursts <- paste0(PATH.TO.DATA, "/", "e_bursts_", thread, artefact,
                                  "_qualitative_analysis_", EPSILON, "_devCEbursts.dat")
  file.name.bursts.emails <- paste0(PATH.TO.DATA, "/", "e_bursts_", thread,
                                   "mbox_", artefact, "_qualitative_analysis_", EPSILON, ".dat")
  file.name.bursts.commit <- paste0(PATH.TO.DATA, "/", "c_bursts_qualitative_analysis_", artefact,
                                    "_", EPSILON, ".dat")

  load(file.name.devCEbursts)
  load(file.name.bursts.commit)
  load(file.name.bursts.emails)

  if(artefact == "features") {
    tagging = "feature"
  } else {
    tagging = "proximity"
  }
  commit.messages <- read.commit.messages.data(paste0(PATH.TO.DATA, "/", "commitMessages_", tagging, ".list"), filter.for.first.line = TRUE)

  results <- mclapply(developer.pairs.c.and.e.bursts, mc.cores=CORES, function(developer.pair){

    if(only.pairs.having.more.than.five.c.e.bursts &&
       (developer.pair$email.bursts <= 5 || developer.pair$commit.bursts <= 5)) {
      result <- list()
      result$mention.rate <- NA
      result$emails.with.matching.subject <- data.frame()
      return(result)
    }
    loginfo(paste(developer.pair$names[[1]], developer.pair$names[[2]]))
    commit.bursts <- burst.commits[[developer.pair$c.number]]
    email.bursts <- burst.emails[[developer.pair$e.number]]

    match.list <- c()
    emails.with.matching.subject <- data.frame()
  
    for(cb in commit.bursts) {
      # get synchronously changed artefacts
      artefacts <- cb$artefacts
      found.match <- FALSE
    
      # get email bursts with a certain time difference
      allowed.ce.burst.distance <- duration(EPSILON, "day")

      # get beginning of commit message of each commit
      current.messages <- commit.messages[commit.messages$commitHash %in% cb$commits$commitHash, "commitMessage"]
      current.messages <- tolower(trimws(current.messages))

      for(eb in email.bursts) {

        # check whether email subject is similar to commit message
	similar.messages <- sub("((((re)|(aw)): *)|(\\[[[:print:]]*\\] *)*)*", "",
				tolower(trimws(eb$emails$subject)))

	#burst.emails.with.matching.subject <- eb$emails[similar.messages %in% tolower(trimws(unlist(current.messages))),]
	burst.emails.with.matching.subject <- eb$emails[sapply(similar.messages, current.messages,
							       FUN=function(mail, commits){
								      if(nchar(mail) == 0) return(FALSE)
								      return(any(startsWith(commits, mail)))
							       }, USE.NAMES=FALSE),]

        emails.with.matching.subject <- rbind(emails.with.matching.subject, burst.emails.with.matching.subject)

	if(nrow(burst.emails.with.matching.subject) > 0 && !found.match && 
           ((min(ymd_hms(eb$emails$creationDate)) <= max(ymd_hms(cb$commits$authorDate))+allowed.ce.burst.distance) &&
           (max(ymd_hms(eb$emails$creationDate)) >= min(ymd_hms(cb$commits$authorDate))-allowed.ce.burst.distance))) {
            loginfo(paste("CE-burst message/subject match:", developer.pair$c.number, developer.pair$e.number))
            found.match = TRUE

        }
      }

      match.list <- c(match.list, found.match)
    }

    result <- list()
    result$mention.rate <- sum(match.list) / length(match.list)
    loginfo(paste("Result:", developer.pair$names[[1]], developer.pair$names[[2]], "message/subject mention-rate:", result))
    result$emails.with.matching.subject <- emails.with.matching.subject
    return(result) 
  })

  if(only.pairs.having.more.than.five.c.e.bursts) {
    output.path.results <- paste0(PATH.TO.DATA, "/", "automaticEvaluation_qualCE_moreThan5CE_",
                                  thread, artefact, "_", EPSILON, ".dat" )
    output.path.filtered.emails <- paste0(PATH.TO.DATA, "/", "emails_filteredOnSubject_moreThan5CE_",
                                  thread, artefact, "_", EPSILON, ".list" )
  } else {
    output.path.results <- paste0(PATH.TO.DATA, "/", "automaticEvaluation_qualCE_",
                                  thread, artefact, "_", EPSILON, ".dat" )
    output.path.filtered.emails <- paste0(PATH.TO.DATA, "/", "emails_filteredOnSubject_",
                                  thread, artefact, "_", EPSILON, ".list" )
  }

  mention.rate <- mclapply(as.list(1:length(developer.pairs.c.and.e.bursts)), mc.cores=CORES, FUN=function(i, results){return(results[[i]]$mention.rate)}, results=results)
  mention.rate <- unlist(mention.rate)

  emails.with.matching.subject <- mclapply(as.list(1:length(developer.pairs.c.and.e.bursts)), mc.cores=CORES, FUN=function(i, results){return(results[[i]]$emails.with.matching.subject)}, results=results)
  emails.with.matching.subject <- rbind.fill(emails.with.matching.subject)


  save(mention.rate, file=output.path.results)
  
  write.table(emails.with.matching.subject, file=output.path.filtered.emails, col.names=FALSE, row.names=FALSE, quote=FALSE, sep="; ")

}

