## Copyright (c) Thomas Bock, 2017
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(parallel)) # tools for parallel execution


## / / / / / / / / / / / / / / / / / / / / / / / / / /
## Response Times in Synchronous Commits on Artifacts
## / / / / / / / / / / / / / / / / / / / / / / / / / /

determine.response.times.in.commits.on.artifacts <- function(commits, analyzeFiles=FALSE, output.path.dat) {

  # check whether this analysis is file-based or not
  artefact.column <- get.artefact.column(analyzeFiles)

  developers <- unique(commits[,"author"])

  pairs.of.developers.data.frame <- as.data.frame(t(combn(developers, 2, simplify=TRUE)))
  colnames(pairs.of.developers.data.frame) <- c("d_a", "d_b")

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))
  # pairs.of.developers <- split(pairs.of.developers.data.frame, seq(1, nrow(pairs.of.developers.data.frame)))

  # if(PARALLEL.JOBS > 1) {
  #   pairs.of.developers.data.frame <- split(pairs.of.developers.data.frame, cut(seq(1,length(pairs.of.developers)), PARALLEL.JOBS))[[JOB.NUMBER]]
  #   pairs.of.developers <- split(pairs.of.developers, cut(seq(1,length(pairs.of.developers)), PARALLEL.JOBS))[[JOB.NUMBER]]
  #   loginfo(paste("Developer pairs to analyze in this job: ", length(pairs.of.developers)))
  # }

  # apply response time determination for each pair of developers
  list.of.response.times.per.commits.on.artifacts <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    commits.d_a <- commits[commits["author"] == pair[1],]
    commits.d_b <- commits[commits["author"] == pair[2],]

    commits.d_a.both.commit.to.artefact <- commits.d_a[commits.d_a[,artefact.column] %in% unique(commits.d_b[,artefact.column]),]

    response.times.in.secs <- c()

    # check whether the two developers work on the same artefact anytime. If not, there can not be a burst between them.
    if(nrow(commits.d_a.both.commit.to.artefact) > 0) {
      commits.d_b.both.commit.to.artefact <- (commits.d_b[commits.d_b[,artefact.column] %in% unique(commits.d_a.both.commit.to.artefact[,artefact.column]),])

      dates.commits.d_a.both.commit.to.artefact <- unique(commits.d_a.both.commit.to.artefact[["authorDate"]])
      dates.commits.d_b.both.commit.to.artefact <- unique(commits.d_b.both.commit.to.artefact[["authorDate"]])

      dates.both.commit.to.artefact.all <- c(dates.commits.d_a.both.commit.to.artefact, dates.commits.d_b.both.commit.to.artefact)
      dates.all.sorted <- sort(dates.both.commit.to.artefact.all)

      # determine response times for each pair of subsequent e-mails of the developer pair in the thread
      response.times.in.secs <- as.numeric(abs(difftime(head(dates.all.sorted,-1), tail(dates.all.sorted,-1), units="secs")))
    }

    return(response.times.in.secs)
  })

  # return a list which containes a vector for each pair of developers,
  # the elements of these vectors are the response times on artifacts for the corresponding developer pair
  save(list.of.response.times.per.commits.on.artifacts, file=output.path.dat)
  return(list.of.response.times.per.commits.on.artifacts)
}

## / / / / / / / / / / / / / / / / / / / / / / / / / /
## Response Times in Email Threads
## / / / / / / / / / / / / / / / / / / / / / / / / / /

determine.response.times.in.threads <- function(emails, output.path.dat) {

  threads <- unique(emails[,"threadId"])
  threads.list <- as.list(threads)

  # for each thread
  list.of.response.times.per.thread <- mclapply(threads.list, mc.cores=CORES, emails=emails, FUN=function(thread, emails){

    # determine emails in this thread
    emails.thread <- emails[emails[["threadId"]]==thread,]

    # get authors in this thread
    authors.thread <- unique(emails.thread[,"author"])

    # check whether at least two authors are participating in this thread
    if(length(authors.thread) < 2) {
      list.of.response.times <- list()
    } else {

      # construct pairs of authors in this thread
      pairs.of.authors.data.frame <- as.data.frame(t(combn(authors.thread, 2, simplify=TRUE)))
      colnames(pairs.of.authors.data.frame) <- c("d_a", "d_b")

      emails.per.author <- lapply(as.vector(authors.thread), FUN=function(author){
        return(emails.thread[emails.thread["author"] == author,])
      })

      # # call garbage collection
      # rm(emails)
      # # gc()

      pairs.of.authors <- as.list(as.data.frame(t(pairs.of.authors.data.frame)))

      # for each pair of developers participating in the thread
      list.of.response.times <- lapply(pairs.of.authors, function(pair){
        emails.d_a <- get.emails.by.author(pair[1], emails.per.author, authors.thread) #emails.thread[emails.thread["author"] == pair[1],]
        emails.d_b <- get.emails.by.author(pair[2], emails.per.author, authors.thread) #emails.thread[emails.thread["author"] == pair[2],]

        email.dates.d_a <- emails.d_a[["creationDate"]]
        email.dates.d_b <- emails.d_b[["creationDate"]]

        email.dates.all <- c(email.dates.d_a, email.dates.d_b)
        email.dates.all.sorted <- sort(email.dates.all)

        # determine response times for each pair of subsequent e-mails of the developer pair in the thread
        response.times.in.secs <- as.numeric(abs(difftime(head(email.dates.all.sorted,-1), tail(email.dates.all.sorted,-1), units="secs")))
        return(response.times.in.secs)
      })
    }

    return(list.of.response.times)
  })

  # return a list which containes a list for each thread,
  # each of these lists in turn contains a vector for each participating developer pair,
  # the elements of these vectors are the response times for the corresponding developer pair in the corresponding thread
  save(list.of.response.times.per.thread, file=output.path.dat)
  return(list.of.response.times.per.thread)
}


## / / / / / / / / / / / / / / / / / / / / / / / / / /
## Response Times in E-Mails
## / / / / / / / / / / / / / / / / / / / / / / / / / /

determine.response.times.in.emails <- function(emails, output.path.dat) {

  authors <- unique(emails[,"author"])

  pairs.of.authors.data.frame <- as.data.frame(t(combn(authors, 2, simplify=TRUE)))
  colnames(pairs.of.authors.data.frame) <- c("d_a", "d_b")

  pairs.of.authors <- as.list(as.data.frame(t(pairs.of.authors.data.frame)))
  # pairs.of.authors <- split(pairs.of.authors.data.frame, seq(1, nrow(pairs.of.authors.data.frame)))

  # if(PARALLEL.JOBS > 1) {
  #   pairs.of.authors.data.frame <- split(pairs.of.authors.data.frame, cut(seq(1,length(pairs.of.authors)), PARALLEL.JOBS))[[JOB.NUMBER]]
  #   pairs.of.authors <- split(pairs.of.authors, cut(seq(1,length(pairs.of.authors)), PARALLEL.JOBS))[[JOB.NUMBER]]
  #   loginfo(paste("Author pairs to analyze in this job: ", length(pairs.of.authors)))
  # }

  # apply response time determination for each pair of authors
  list.of.response.times <- mclapply(pairs.of.authors, mc.cores=CORES, FUN=function(pair){
    emails.d_a <- emails[emails["author"] == pair[1],]
    emails.d_b <- emails[emails["author"] == pair[2],]

    response.times.in.secs <- c()

    dates.emails.d_a <- unique(emails.d_a[["creationDate"]])
    dates.emails.d_b <- unique(emails.d_b[["creationDate"]])

    email.dates.all <- c(dates.emails.d_a, dates.emails.d_b)
    email.dates.all.sorted <- sort(email.dates.all)

    # determine response times for each pair of subsequent e-mails of the considered pair of authors
    response.times.in.secs <- as.numeric(abs(difftime(head(email.dates.all.sorted,-1), tail(email.dates.all.sorted,-1), units="secs")))

    return(response.times.in.secs)
  })

  # return a list which containes a vector for each pair of authors,
  # the elements of these vectors are the response times for the corresponding author pair
  save(list.of.response.times, file=output.path.dat)
  return(list.of.response.times)
}
