## Copyright (c) Thomas Bock, 2016
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(logging)) # logging

## / / / / / / / / / / / / / / / / / / / / / / / /
## Functions to run different types of simulations
## / / / / / / / / / / / / / / / / / / / / / / / /

do.commit.randomization <- function(commits.proximity, output.file.name.proximity, commits.feature, output.file.name.feature) {

  # NOTICE: parameter EPSILON represents the number of simulation to perform!
  simulations <- c(1:EPSILON)

  if(PARALLEL.JOBS > 1) {
    simulations <- split(simulations, cut(seq(1,length(simulations)), PARALLEL.JOBS))[[JOB.NUMBER]]
  }
  loginfo(paste("Simulations to perform in this job: ", length(simulations)))

  for(i in simulations) {

    # create individual subdirectory for each simulation
    path.to.simulation <- file.path(PATH.TO.DATA, paste0("simulation", i))
    if(!file.exists(path.to.simulation)) {
      dir.create(path.to.simulation)
      file.create(file.path(path.to.simulation, "e_proximity.list"))
      loginfo(paste0("Create empty emails-file for commit-simulation ", i))
    }
    loginfo(paste0("Generate simulation ", i, " ..."))

    # generate simulated commit time-series (same time-series for proximity and feature)
    authors <- unique(commits.proximity[["author"]])
    simulation <- mclapply(authors, mc.cores=CORES, FUN=randomize.commit.timestamps, commits.proximity=commits.proximity, commits.feature=commits.feature)
    simulation.proximity <- mclapply(as.list(1:length(authors)), mc.cores=CORES, FUN=function(i, simulation){return(simulation[[i]]$proximity)}, simulation=simulation)
    simulation.feature <- mclapply(as.list(1:length(authors)), mc.cores=CORES, FUN=function(i, simulation){return(simulation[[i]]$feature)}, simulation=simulation)
    simulation.proximity <- rbind.fill(simulation.proximity)
    simulation.feature <- rbind.fill(simulation.feature)

    # write simulated commits to simulation subdirectory
    output.path.proximity <- file.path(path.to.simulation, output.file.name.proximity)
    output.path.feature <- file.path(path.to.simulation, output.file.name.feature)
    write.table(simulation.proximity, file=output.path.proximity, col.names=FALSE, row.names=FALSE, quote=FALSE, sep="; ")
    loginfo(paste0("Wrote commit-simulation ", i, " to ", output.path.proximity))
    write.table(simulation.feature, file=output.path.feature, col.names=FALSE, row.names=FALSE, quote=FALSE, sep="; ")
    loginfo(paste0("Wrote commit-simulation ", i, " to ", output.path.feature))
  }

  loginfo(paste("Finished generation of", length(simulations), "simulations."))
}

randomize.commit.timestamps <- function(author, commits.proximity, commits.feature) {

  ## RANDOMIZE COMMITS BASED ON PROXIMITY DATA

  # get all commits of one author
  commits.of.author <-get.commits.of.author(author, commits.proximity)

  unsorted.time.stamps <- commits.of.author[["authorDate"]]
  sorted.time.stamps <- sort(unique(commits.of.author[["authorDate"]]))

  start.time.stamp <- sorted.time.stamps[1]
  end.time.stamp <- sorted.time.stamps[length(sorted.time.stamps)]

  time.intervals <- int_diff(sorted.time.stamps)

  intervals.lengths <- determine.length.of.interval.in.seconds(time.intervals)

  # randomly swap time-differences between two subsequent commits
  random.numbers <- sample(c(1:length(intervals.lengths)), length(intervals.lengths), replace=FALSE)
  randomized.intervals.lengths <- intervals.lengths[random.numbers]

  seconds.from.start <- c(0,cumsum(randomized.intervals.lengths))

  randomized.time.stamps <- ymd_hms(rep(start.time.stamp, length(sorted.time.stamps))) + seconds(seconds.from.start)

  # end time-stamp should be the same as before!
  # print(ymd_hms(randomized.time.stamps[length(randomized.time.stamps)])==ymd_hms(end.time.stamp))

  # replace commit dates
  new.commit.times <- randomized.time.stamps[mapply(unsorted.time.stamps, FUN=function(original.time.stamp) {which(sorted.time.stamps==original.time.stamp)})]
  commits.of.author[["authorDate"]] <- new.commit.times


  ## APPLY RANDOMIZED COMMIT DATES TO FEATURE DATA

  # get all commits of one author
  commits.of.author.feature <- get.commits.of.author(author, commits.feature)
  unsorted.time.stamps.feature <- commits.of.author.feature[["authorDate"]]

  # replace commit dates
  new.commit.times.feature <- randomized.time.stamps[mapply(unsorted.time.stamps.feature, FUN=function(original.time.stamp) {which(sorted.time.stamps==original.time.stamp)})]
  commits.of.author.feature[["authorDate"]] <- new.commit.times.feature


  # RETURN RANDOMIZED PROXIMITY AND FEATURE COMMITS
  new.commits <- list()
  new.commits$proximity <- commits.of.author
  new.commits$feature <- commits.of.author.feature

  return(new.commits)
}

do.email.randomization <- function(emails, output.file.name) {

  # NOTICE: parameter EPSILON represents the number of simulation to perform!
  simulations <- c(1:EPSILON)

  if(PARALLEL.JOBS > 1) {
    simulations <- split(simulations, cut(seq(1,length(simulations)), PARALLEL.JOBS))[[JOB.NUMBER]]
  }
  loginfo(paste("Simulations to perform in this job: ", length(simulations)))

  for(i in simulations) {

    # create individual subdirectory for each simulation
    path.to.simulation <- file.path(PATH.TO.DATA, paste0("simulation", i))
    if(!file.exists(path.to.simulation)) {
      dir.create(path.to.simulation)
    }
    loginfo(paste0("Generate simulation ", i, " ..."))

    # generate simulated email time-series (individual simulation for each pair of authors)
    authors <- unique(emails[["author"]])
    pairs.of.authors.data.frame <- as.data.frame(t(combn(authors, 2, simplify=TRUE)))
    pairs.of.authors <- as.list(as.data.frame(t(pairs.of.authors.data.frame)))

    email.simulation <- mclapply(pairs.of.authors, mc.cores=CORES, FUN=randomize.email.timestamps, emails=emails)

    simulated.email.data <- list()
    simulated.email.data$pairs.of.authors <- pairs.of.authors.data.frame
    simulated.email.data$simulations <- email.simulation

    # remove unnecessary columns in order to reduce memory consumption
    simulated.email.data$simulations <- lapply(simulated.email.data$simulations, FUN=function(x) {x[["authorEmail"]] <- NULL; x[["subject"]] <- NULL; x[["messageId"]] <- NULL; return(x)})



    # write simulated email time-series to simulation subdirectory
    output.path <- file.path(path.to.simulation, output.file.name)
    save(simulated.email.data, file=output.path)
    loginfo(paste0("Saved email-simulation ", i, " to ", output.path))
  }

  loginfo(paste("Finished generation of", length(simulations), "simulations."))
}

randomize.email.timestamps <- function(pair.of.authors, emails) {

  ## RANDOMIZE EMAIL TIME-SERIES FOR PAIR OF AUTHORS

  # get all emails of the pair
  emails.of.author1 <- get.emails.of.author(pair.of.authors[1], emails)
  emails.of.author2 <- get.emails.of.author(pair.of.authors[2], emails)
  emails.of.both.authors <- rbind(emails.of.author1, emails.of.author2)

  unsorted.time.stamps <- emails.of.both.authors[["creationDate"]]
  sorted.time.stamps <- sort(unique(emails.of.both.authors[["creationDate"]]))

  start.time.stamp <- sorted.time.stamps[1]
  end.time.stamp <- sorted.time.stamps[length(sorted.time.stamps)]

  time.intervals <- int_diff(sorted.time.stamps)

  intervals.lengths <- determine.length.of.interval.in.seconds(time.intervals)

  # randomly swap time-differences between two subsequent emails
  random.numbers <- sample(c(1:length(intervals.lengths)), length(intervals.lengths), replace=FALSE)
  randomized.intervals.lengths <- intervals.lengths[random.numbers]

  seconds.from.start <- c(0,cumsum(randomized.intervals.lengths))

  randomized.time.stamps <- ymd_hms(rep(start.time.stamp, length(sorted.time.stamps))) + seconds(seconds.from.start)

  # end time-stamp should be the same as before!
  # print(ymd_hms(randomized.time.stamps[length(randomized.time.stamps)])==ymd_hms(end.time.stamp))

  # replace email dates
  new.email.times <- randomized.time.stamps[mapply(unsorted.time.stamps, FUN=function(original.time.stamp) {which(sorted.time.stamps==original.time.stamp)})]
  emails.of.both.authors[["creationDate"]] <- new.email.times

  # RETURN RANDOMIZED EMAILS
  return(emails.of.both.authors)
}
