## Copyright (c) Thomas Bock, 2016-2018
## bockthom@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(logging)) # logging
suppressWarnings(library(stats))

library(smoother)
library(dtw)
library(ggplot2)

PLOTS.DIRECTORY <- "/curve-plots"

INCL_LESS_5 <- TRUE

BW <- 2 # 1, 2, 3

## / / / / / / / / / / / / / / / / / / / / / / / /
## Functions to generate smoothed curves
## / / / / / / / / / / / / / / / / / / / / / / / /

do.smoothing <- function(analysis.type, analysis.file.name.part, commits, analyze.files, analysis.type.email, simulated=FALSE, thread.based=FALSE,
                based.on.commit.messages=FALSE, emails) {
  path.to.c.burst.data <- paste0(PATH.TO.DATA, "/", "c_bursts_", analysis.file.name.part ,"_", EPSILON, ".dat")
  path.to.c.burst.analysis <- paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", analysis.file.name.part,"_", EPSILON, ".dat")

  file.name.part <- ""
  output.name.part <- ""

  if(INCL_LESS_5) output.name.part <- "inclLess5_"

  if(based.on.commit.messages) {
    file.name.part <- paste0("commitMessages_", analysis.file.name.part, "_")
    output.name.part <- paste0("commitMessages_", output.name.part)
  }


  if(thread.based) {
    path.to.e.burst.data <- paste0(PATH.TO.DATA, "/", "e_bursts_thread_", file.name.part, EPSILON, ".dat")
    path.to.e.burst.analysis <- paste0(PATH.TO.DATA, "/", "e_bursts_thread_", file.name.part, "analysis_", EPSILON, ".dat")
    threads="thread_"
  } else {
    path.to.e.burst.data <- paste0(PATH.TO.DATA, "/", "e_bursts_", file.name.part, EPSILON, ".dat")
    path.to.e.burst.analysis <- paste0(PATH.TO.DATA, "/", "e_bursts_", file.name.part, "analysis_", EPSILON, ".dat")
    threads=""
  }

  # Remove commits on file level resp. base feature
  if(analysis.type=="function") {
    analyzed.commits <- remove.file.level.commits(commits)
  } else if(analysis.type=="feature") {
    analyzed.commits <- remove.base.feature.commits(commits)
  } else {
    analyzed.commits <- commits
  }

  ## LOAD RELEVANT DATA

  loginfo(paste("Load file:", path.to.c.burst.data))
  load(path.to.c.burst.data)
  c.bursts <- bursts
  loginfo(paste("Load file:", path.to.c.burst.analysis))
  load(path.to.c.burst.analysis)
  c.bursts.meta.data <- data

  if(!simulated){
    loginfo(paste("Load file:", path.to.e.burst.data))
    load(path.to.e.burst.data)
    e.bursts <- bursts
    loginfo(paste("Load file:", path.to.e.burst.analysis))
    load(path.to.e.burst.analysis)
    e.bursts.meta.data <- data
  } else {
    if(startsWith(CASESTUDY, "qemu_fixed")) {
      loginfo(paste("Load file:", paste0(PATH.TO.DATA, "/e_simulation_short.dat")))
      load(paste0(PATH.TO.DATA, "/e_simulation_short.dat"))
    } else {
      loginfo(paste("Load file:", paste0(PATH.TO.DATA, "/e_simulation.dat")))
      load(paste0(PATH.TO.DATA, "/e_simulation.dat"))
    }
    emails <- simulated.email.data$simulations
    loginfo(paste("Load file:", paste0(PATH.TO.DATA, "/e_bursts_", threads, file.name.part, EPSILON, ".dat")))
    load(paste0(PATH.TO.DATA, "/e_bursts_", threads, file.name.part, EPSILON, ".dat"))
    e.bursts <- bursts
    loginfo(paste("Load file:", paste0(PATH.TO.DATA, "/e_bursts_", threads, file.name.part, "analysis_", EPSILON, ".dat")))
    load(paste0(PATH.TO.DATA, "/e_bursts_", threads, file.name.part, "analysis_", EPSILON, ".dat"))
    e.bursts.meta.data <- data
  }

  pairs.of.developers.data.frame <- c.bursts[[1]]

  # Get all pairs of developers which have more than 5 c-bursts
  if(!INCL_LESS_5){
    pairs.of.developers.data.frame <- pairs.of.developers.data.frame[which(c.bursts.meta.data$bursts.per.pair > 5),]
  } else {
    pairs.of.developers.data.frame <- pairs.of.developers.data.frame[which(c.bursts.meta.data$bursts.per.pair > 0),]
  }

  pairs.of.authors.data.frame <- e.bursts[[1]]

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  # For all developers with more than 5 c-bursts: Check if they also have more than 5 e-bursts
  for(pair in pairs.of.developers) {

    # Get regarding e-bursts
    email.pair.no <- get.pair.no(pair[1], pair[2], pairs.of.authors.data.frame)

    if(length(email.pair.no) != 1) {
      email.pair.no <- get.pair.no(pair[2], pair[1], pairs.of.authors.data.frame)
    }
    if(length(email.pair.no) != 1) {
      # The current pair did not write emails at all. Therefore, remove the pair from the list of pairs to consider...
      pairs.of.developers.data.frame <- pairs.of.developers.data.frame[(rownames(pairs.of.developers.data.frame)
                                                                        != get.pair.no(pair[1], pair[2],pairs.of.developers.data.frame)),]
    } else {
      if(email.pair.no > length(e.bursts.meta.data$bursts.per.pair) || (!INCL_LESS_5 && e.bursts.meta.data$bursts.per.pair[[email.pair.no]] <= 5) || (INCL_LESS_5 && e.bursts.meta.data$bursts.per.pair[[email.pair.no]] <= 0)) {
        # The current pair does not have more than 5 e-bursts. Therefore, remove the pair form the list of pairs to consider...
        pairs.of.developers.data.frame <- pairs.of.developers.data.frame[(rownames(pairs.of.developers.data.frame)
                                                                          != get.pair.no(pair[1], pair[2],pairs.of.developers.data.frame)),]
      }
    }
  }

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  if(!INCL_LESS_5) {
    loginfo(paste("Pairs with more than 5 c-bursts and more than 5 e-bursts:", length(pairs.of.developers)))
  }

  band.window=(BW*EPSILON)
  # Check if there are pairs with more than 5 c-bursts and more than 5 e-bursts
  if(length(pairs.of.developers) > 0) {

    sigma=(BW*EPSILON)

    ## Perform smoothing!
    loginfo("Generate smoothed c-curves...")
    smoothed.c.curves <- smooth.c(commits, c.bursts[[2]], pairs.of.developers.data.frame, pairs.of.authors.data.frame, e.bursts[[2]], analyze.files, sigma=sigma)
    #stop()
    loginfo("Generate smoothed e-curves...")
    smoothed.e.curves <- smooth.e(emails, e.bursts[[2]], pairs.of.authors.data.frame, pairs.of.developers.data.frame, c.bursts[[2]], simulated=simulated, thread.based=thread.based, sigma=sigma)

    # calculate correlation coefficients
    dtw.distances <- vector(mode="list", length(pairs.of.developers))

    for(i in c(1:length(pairs.of.developers))) {
      dtw.distances[[i]] <- calculate.dtw(smoothed.c.curves[[i]], smoothed.e.curves[[i]], analysis.file.name.part, i, thread.based=thread.based, band.window=band.window, sigma=sigma, simulated=simulated)
    }
    #stop("Only generate plots")

    print("=========DTW distances===========")
    print("Mean:")
    print(mean(unlist(dtw.distances)))
    print("Median:")
    print(median(unlist(dtw.distances)))

    # save results to file
    data <- list()
    data$compared.pairs.data.frame <- pairs.of.developers.data.frame
    data$smoothed.c.curves <- smoothed.c.curves
    data$smoothed.e.curves <- smoothed.e.curves
    data$dtw.distances <- dtw.distances

    output.path <- paste0(PATH.TO.DATA, "/", "smoothed_curves_", threads, analysis.file.name.part ,"_", output.name.part, EPSILON, "_SigmaAsBandWindow_bandWindowSize", band.window, ".dat")
    if(simulated) output.path <- paste0(PATH.TO.DATA, "/", "simulated_smoothed_curves_", threads, analysis.file.name.part ,"_", output.name.part, EPSILON, "_SigmaAsBandWindow_bandWindowSize", band.window, ".dat")
      loginfo(paste("Save smoothed curves and their DTW distances to file:", output.path))
      save(data, file = output.path)

  } else {
    if(!INCL_LESS_5) {
      loginfo("There are no pairs with more than 5 c-bursts and more than 5 e-bursts!")
    } else {
      loginfo("There are no pairs having c- and e-bursts!")
    }

    output.path <- paste0(PATH.TO.DATA, "/", "smoothed_curves_", threads, analysis.file.name.part ,"_", output.name.part, EPSILON, ".dat")
    if(simulated) output.path <- paste0(PATH.TO.DATA, "/", "simulated_smoothed_curves_", threads, analysis.file.name.part ,"_", output.name.part, EPSILON, "_SigmaAsBandWindow_bandWindowSize", band.window, ".dat")

    loginfo(paste0("Save empty file: ", output.path))
    data <- list()
    data$compared.pairs.data.frame <- data.frame()
    data$smoothed.c.curves <- c()
    data$smoothed.e.curves <- c()
    data$dtw.distances <- c()
    save(data, file=output.path)
  }
}

smooth.c <- function(commits, commit.bursts, pairs.of.developers.data.frame, pairs.of.authors.data.frame, email.bursts, analyzeFiles, sigma) {

  # check whether this analysis is file-based or not
  artefact.column <- get.artefact.column(analyzeFiles)

  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  i=0

  # Iterate over all pairs of developers
  list.of.smoothed.c.curves <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    i<<-i+1
    {
      # Get bursts
      pair.no <- get.pair.no(pair[1], pair[2], pairs.of.developers.data.frame)
      bursts.data.frame <- commit.bursts[[pair.no]]

      activities1 <- get.commits.of.author(pair[1], commits)
      activities2 <- get.commits.of.author(pair[2], commits)

      ## calculate upper and lower (from both c- and e-bursts) for the current pair of developers

      # Get corresponding email bursts
      email.pair.no <- get.pair.no(pair[1], pair[2], pairs.of.authors.data.frame)
      if(length(email.pair.no) != 1) {
        email.pair.no <- get.pair.no(pair[2], pair[1], pairs.of.authors.data.frame)
      }
      email.bursts.data.frame <- email.bursts[[email.pair.no]]
      lower <- min(as.numeric(as.Date(bursts.data.frame[,1])), as.numeric(as.Date(email.bursts.data.frame[,1])))
      upper <- max(as.numeric(as.Date(bursts.data.frame[,2])), as.numeric(as.Date(email.bursts.data.frame[,2])))

      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      smoothed.curve <- vector(mode="numeric", length=length(c(lower:upper)))
      names(smoothed.curve) <- c(lower:upper) #c(as.numeric(as.Date(START.DATE)):as.numeric(as.Date(END.DATE))))
      j=0
      max.individual <- 1
      for(interval in bursts) {
        j=j+1
        burst.activities1 <- activities1[(activities1[["authorDate"]] >= interval[1] & activities1[["authorDate"]]<= interval[2]),]
        burst.activities2 <- activities2[(activities2[["authorDate"]] >= interval[1] & activities2[["authorDate"]] <= interval[2]),]

        # Get the artefacts the other developer committed to within the burst time-interval and check for overlapping artefacts
        artefacts.commited.to.by.both <- intersect(unique(burst.activities1[[artefact.column]]),unique(burst.activities2[[artefact.column]]))

        burst.activities1 <- burst.activities1[burst.activities1[[artefact.column]] %in% artefacts.commited.to.by.both,]
        burst.activities2 <- burst.activities2[burst.activities2[[artefact.column]] %in% artefacts.commited.to.by.both,]

        burst.activity.times.both <- c(burst.activities1[["authorDate"]],burst.activities2[["authorDate"]])

        burst.activities.per.day <- aggregate(burst.activity.times.both ~ as.Date(times), data.frame(times=burst.activity.times.both), FUN=length)

        SIGMA <- sigma

        peak <- as.numeric(as.Date(interval[3]))

        zeta <- peak-as.numeric(as.Date(interval[1]))

        smoothing.start.all <- lower
        smoothing.end.all <- upper
        smoothing.start.individual <- (as.numeric(as.Date(interval[1]))-SIGMA)
        smoothing.end.individual <- (as.numeric(as.Date(interval[2]))+SIGMA)

        vector.to.smooth <- vector(mode="numeric", length=length(c(smoothing.start.individual:smoothing.end.individual)))
        names(vector.to.smooth) <- c(smoothing.start.individual:smoothing.end.individual)

        for(a in c(1:nrow(burst.activities.per.day))) {
          if(as.numeric(as.Date(burst.activities.per.day[a,1])) %in% names(vector.to.smooth)) {

            vector.to.smooth[names(vector.to.smooth)==as.numeric(as.Date(burst.activities.per.day[a,1]))] <- (burst.activities.per.day[a,2] )
          }
        }

        smoothed.burst <- perform.smoothing(vector.to.smooth, SIGMA)
        smoothed.burst[is.na(smoothed.burst)] <- 0

        names(smoothed.burst) <- names(vector.to.smooth)
        smoothed.burst <- smoothed.burst[names(smoothed.burst) %in% c(smoothing.start.all:smoothing.end.all)]

        smoothed.curve <- tapply(c(smoothed.curve,smoothed.burst), names(c(smoothed.curve,smoothed.burst)), sum)

        smoothed.curve <- smoothed.curve[order(as.numeric(names(smoothed.curve)))]
        max.individual <- max(smoothed.burst)
      }

      smoothed.curve[smoothed.curve > max.individual] <- max.individual
      loginfo(paste("Finished smoothing c-curve of", pair[1], pair[2]))

      return(smoothed.curve)
    }
  })

  return(list.of.smoothed.c.curves)
}

smooth.e <- function(emails, email.bursts, pairs.of.authors.data.frame, pairs.of.developers.data.frame, commit.bursts, simulated, thread.based, sigma) {
  i <- 0
  pairs.of.developers <- as.list(as.data.frame(t(pairs.of.developers.data.frame)))

  # Iterate over all pairs of developers
  list.of.smoothed.e.curves <- mclapply(pairs.of.developers, mc.cores=CORES, FUN=function(pair){
    i<<-i+1
    #if(i==25)
    {

      # Get corresponding email bursts
      pair.no <- get.pair.no(pair[1], pair[2], pairs.of.authors.data.frame)
      if(length(pair.no) != 1) {
        pair.no <- get.pair.no(pair[2], pair[1], pairs.of.authors.data.frame)
      }
      bursts.data.frame <- email.bursts[[pair.no]]

      # check whether this analysis is on simulated data or not
      if(!simulated) {
        activities1 <- get.emails.of.author(pair[1], emails)
        activities2 <- get.emails.of.author(pair[2], emails)
        activities <- rbind(activities1,activities2)
      } else {
        activities <- emails[[pair.no]]
        activities1 <- activities[activities[["author"]]==pair[1],]
        activities2 <- activities[activities[["author"]]==pair[2],]
      }

      # calculate upper and lower (from both c- and e-bursts) for the current pair of developers
      commit.bursts.data.frame <- commit.bursts[[get.pair.no(pair[1], pair[2], pairs.of.developers.data.frame)]]
      lower <- min(as.numeric(as.Date(bursts.data.frame[,1])), as.numeric(as.Date(commit.bursts.data.frame[,1])))
      upper <- max(as.numeric(as.Date(bursts.data.frame[,2])), as.numeric(as.Date(commit.bursts.data.frame[,2])))

      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      smoothed.curve <- vector(mode="numeric", length=length(c(lower:upper)))
      names(smoothed.curve) <- c(lower:upper) #vector(mode="numeric", length=c(as.numeric(as.Date(START.DATE)):as.numeric(as.Date(END.DATE))))
      j=0
      max.individual <- 1
      for(interval in bursts) {
        j=j+1

        if(!thread.based) {
          burst.activities <- activities[(activities[["creationDate"]] >= interval[1] & activities[["creationDate"]]<= interval[2]),]

          burst.activity.times.both <- burst.activities[["creationDate"]]

        } else { # thread based
          burst.activities1 <- activities1[(activities1[["creationDate"]] >= interval[1] & activities1[["creationDate"]]<= interval[2]),]
          burst.activities2 <- activities2[(activities2[["creationDate"]] >= interval[1] & activities2[["creationDate"]] <= interval[2]),]

          # Get the threads the other developer replied to within the burst time-interval and check for overlapping threads
          threads.replied.to.by.both <- intersect(unique(burst.activities1[["threadId"]]),unique(burst.activities2[["threadId"]]))

          burst.activities1 <- burst.activities1[burst.activities1[["threadId"]] %in% threads.replied.to.by.both,]
          burst.activities2 <- burst.activities2[burst.activities2[["threadId"]] %in% threads.replied.to.by.both,]

          burst.activity.times.both <- c(burst.activities1[["creationDate"]],burst.activities2[["creationDate"]])
        }

        burst.activity.times.both <- sort(burst.activity.times.both)
        burst.activities.per.day <- aggregate(burst.activity.times.both ~ as.Date(times), data.frame(times=burst.activity.times.both), FUN=length)

        SIGMA <- sigma

        peak <- as.numeric(as.Date(interval[3]))

        zeta <- peak-as.numeric(as.Date(interval[1]))

        smoothing.start.all <- lower
        smoothing.end.all <- upper
        smoothing.start.individual <- (as.numeric(as.Date(interval[1]))-SIGMA)
        smoothing.end.individual <- (as.numeric(as.Date(interval[2]))+SIGMA)

        vector.to.smooth <- vector(mode="numeric", length=length(c(smoothing.start.individual:smoothing.end.individual)))
        names(vector.to.smooth) <- c(smoothing.start.individual:smoothing.end.individual)

        for(a in c(1:nrow(burst.activities.per.day))) {
          if(as.numeric(as.Date(burst.activities.per.day[a,1])) %in% names(vector.to.smooth)) {
            vector.to.smooth[names(vector.to.smooth)==as.numeric(as.Date(burst.activities.per.day[a,1]))] <- (burst.activities.per.day[a,2])
          }
        }

        smoothed.burst <- perform.smoothing(vector.to.smooth, SIGMA)
        smoothed.burst[is.na(smoothed.burst)] <- 0
        names(smoothed.burst) <- names(vector.to.smooth)


        smoothed.burst <- smoothed.burst[names(smoothed.burst) %in% c(smoothing.start.all:smoothing.end.all)]

        smoothed.curve <- tapply(c(smoothed.curve,smoothed.burst), names(c(smoothed.curve,smoothed.burst)), sum)

        smoothed.curve <- smoothed.curve[order(as.numeric(names(smoothed.curve)))]
        max.individual <- max(smoothed.burst)
      }

      smoothed.curve[smoothed.curve > max.individual] <- max.individual

      loginfo(paste("Finished smoothing e-curve of", pair[1], pair[2]))
      return(smoothed.curve)
    }
  })

  return(list.of.smoothed.e.curves)
}

calculate.dtw <- function(c.curve, e.curve, analysis.file.name.part, i, thread.based=FALSE, band.window, sigma, simulated) {
  upper <- max(as.numeric(names(c.curve)), na.rm=TRUE)
  lower <- min(as.numeric(names(c.curve)), na.rm=TRUE)

  if(thread.based) {
    threads = "threads_"
    threads.printable = "threads"
  } else {
    threads = ""
    threads.printable = ""
  }

  SIGMA <- sigma


  if(simulated) print("SIMULATION")

  if(!simulated){
    dir.create(file.path(PATH.TO.DATA, PLOTS.DIRECTORY, "rectangular_smoothed_curves"), showWarnings = FALSE, recursive = TRUE) # "curves", "rectangular_smoothed_curves"

    df.e <- data.frame(days=c(rep(c(1:ceiling(0.25*length(e.curve))),4))[1:length(e.curve)], activity=e.curve, groups=c(rep(1,ceiling(0.25*length(e.curve))),rep(2,ceiling(0.25*length(e.curve))), rep(3,ceiling(0.25*length(e.curve))), rep(4,ceiling(0.25*length(e.curve))-(ceiling(0.25*length(e.curve))*4-length(e.curve)))))

    df.c <- data.frame(days=c(rep(c(1:ceiling(0.25*length(e.curve))),4))[1:length(e.curve)], activity=c.curve, groups=c(rep(1,ceiling(0.25*length(e.curve))),rep(2,ceiling(0.25*length(e.curve))), rep(3,ceiling(0.25*length(e.curve))), rep(4,ceiling(0.25*length(e.curve))-(ceiling(0.25*length(e.curve))*4-length(e.curve)))))
    pdf(file=paste0(PATH.TO.DATA, PLOTS.DIRECTORY,"/rectangular_smoothed_curves/", "extended_rectangular_smoothed_curves_", threads, analysis.file.name.part ,"_", EPSILON,"_", i, "_sigma", SIGMA, ".pdf")) # "extended_rectangular_smoothed_curves_" # "extended_curves_"
    plot <- ggplot(df.e, aes(x=days,y=activity,group=1)) + geom_line(aes(color="E-curve"))
    plot <- plot + geom_line(data=df.c, aes(x=days, y=activity,group=2, color="C-curve"))
    plot <- plot + ggtitle(paste("Rectangular smoothed curves", threads.printable, analysis.file.name.part, ", time-window", EPSILON, ", pair", i, ", sigma=", SIGMA)) #"Rectangular smoothed curves" # "Curves (without smoothing)"
    plot <- plot + scale_colour_manual(name="curves", values=c("E-curve"="red","C-curve"="black")) + theme(legend.position="bottom")
    plot <- plot + facet_grid(groups~.)
    print(plot)
    dev.off()
  }

 if(!is.null(c.curve) && !is.null(e.curve)){
   dynamic.time.warping <- dtw(x=c.curve,y=e.curve, window.type="sakoechiba", window.size=band.window)

   if(!simulated){

   dir.create(file.path(PATH.TO.DATA, PLOTS.DIRECTORY, "dtw_rectangular_smoothed"), showWarnings = FALSE, recursive = TRUE) # "curves", "rectangular_smoothed_curves"
   pdf(file=paste0(PATH.TO.DATA, PLOTS.DIRECTORY, "/dtw_rectangular_smoothed/", "DTW_rectangular_smoothed_curves_", threads, analysis.file.name.part ,"_time-window", EPSILON,"_", i, "_sigma", SIGMA, "_bandwindow", band.window, ".pdf"))
   dtwPlot(dynamic.time.warping, xlab="C-curve", ylab="E-curve", xts=c.curve, yts=e.curve, type="threeway", main=paste0("DTW_threeway_rectangular_smoothed_curves_\n", threads, analysis.file.name.part ,", time-window=", EPSILON,", pair", i, ", sigma=", SIGMA), sub=paste("Distance:", dynamic.time.warping$distance, "Band window:", band.window))
   dtwPlot(dynamic.time.warping, xlab="C-curve", ylab="E-curve", xts=c.curve, yts=e.curve, type="twoway", main=paste0("DTW_twoway_rectangular_smoothed_curves_\n", threads, analysis.file.name.part ,", time-window=", EPSILON,", pair", i, ", sigma=", SIGMA), sub=paste("Distance:", dynamic.time.warping$distance, "Band window:", band.window))
   dtwPlot(dynamic.time.warping, xlab="C-curve", ylab="E-curve", type="alignment", main=paste0("DTW_alignment_rectangular_smoothed_curves_\n", threads,  analysis.file.name.part ,", time-window=", EPSILON,", pair", i, ", sigma=", SIGMA), sub=paste("Distance:", dynamic.time.warping$distance, "Band window:", band.window))
   dev.off()
   }

   print(paste("i=",i))
   print(paste("dtw distance=",dynamic.time.warping$distance))
   print(paste("dtw normalized distance=",dynamic.time.warping$normalizedDistance))
   print(dynamic.time.warping)

   return(dynamic.time.warping$distance)
 } else {
   return(1000)
 }
#  return(NULL)
}

perform.smoothing <- function(x,sigma) {
  density(x, kernel="rectangular", bw=sigma, n=length(x))$y
}
