## Copyright (c) Thomas Bock, 2016-2019
## bockthom@fim.uni-passau.de
##
## Copyright (c) Claus Hunsen, 2016
## hunsen@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

# STAND ALONE APPLICATION

## / / / / / / / / / / / / / /
## Set environment for script
## / / / / / / / / / / / / / /

options(stringsAsFactors = FALSE)
options(warn = 1)

# determine path of this very script file
getScriptPath <- function(){
  cmd.args <- commandArgs()
  m <- regexpr("(?<=^--file=).+", cmd.args, perl=TRUE)
  script.dir <- dirname(regmatches(cmd.args, m))
  if(length(script.dir) == 0) stop("can't determine script dir: please call the script with Rscript")
  if(length(script.dir) > 1) stop("can't determine script dir: more than one '--file' argument detected")
  return(script.dir)
}

# source(file.path(getScriptPath(), "util.R"), chdir = TRUE) # get utility functions
# source(file.path(getScriptPath(), "coordinationBursts.R"), chdir = TRUE) # get functions to identify co-commit and email bursts
# source(file.path(getScriptPath(), "analysis.R"), chdir = TRUE) # get functions to run several kinds of analyses
# source(file.path(getScriptPath(), "simulation.R"), chdir = TRUE) # get function to generate simulated time-series

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(logging)) # logging
suppressWarnings(library(matrixStats)) # colMedians
suppressWarnings(library(lsr)) # cohen's d effect size
suppressWarnings(library(effsize)) # cliff's delta effect size
suppressWarnings(library(rcompanion)) # r effect size for paired wilcoxon signed rank test

basicConfig()
#addHandler(writeToConsole)

## / / / / / / / / / / / / / /
## Configuration parameters
## / / / / / / / / / / / / / /

## Read command line parameters
if (!exists("cargs"))
  cargs = commandArgs(trailingOnly = TRUE)

if(length(cargs)==6) {

  # case-study
  CASESTUDY <- cargs[2]

  # directory to read input and write output files
  PATH.TO.DATA <- paste0(cargs[1], CASESTUDY) #"~/MA/qemu"

  # directory to read R object data of simulations
  PATH.TO.SIMULATION.DATA <- file.path(PATH.TO.DATA, "simulation")

  # analysis type (files, features, functions)
  ANALYSIS <- cargs[3]

  # the maximal allowed time difference between two commits or two mails to be considered within a burst
  EPSILON  <- as.numeric(cargs[4])

  THREADS <- as.numeric(cargs[5])
  if(THREADS < 0 || THREADS > 1) stop("Invalid argument: Threads must be 0 or 1")

  SIMULATIONS <- as.numeric(cargs[6])
} else {
  stop(paste("Invalid number of arguments. Please run the script with the following args: \n",
             "% Rscript t-test.R path.to.data casestudy analysis epsilon threads simulations \n", "Set the threads parameter to 1 if you want to analyze thread-based E-bursts, otherwise set this parameter to 0.", sep=" "))
}

loginfo(paste0("CASESTUDY=", CASESTUDY))
loginfo(paste0("EPSILON=", EPSILON))
loginfo(paste0("THREADS=", THREADS))
loginfo(paste0("ANALYSIS=", ANALYSIS))

if(THREADS==0) threads=""
if(THREADS==1) threads="thread_"

CM = FALSE #TRUE #FALSE
COMMIT.MESSAGES = ""
if(CM) COMMIT.MESSAGES = "commitMessages_"

INCL_LESS_5 = TRUE #FALSE
incl_less_5 = ""
if(INCL_LESS_5) incl_less_5 = "inclLess5_"

synchronous.development.t.test <- function(paired=TRUE, median=FALSE, test="t"){

  # load real c-burst data
  load(paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", ANALYSIS,"_", EPSILON, ".dat"))
  real.bursts.per.pair <- data$bursts.per.pair

  # load simulated c-burst data
  simulated.bursts.per.pair <- vector("list", SIMULATIONS)

  for(i in c(1:SIMULATIONS)) {
    load(paste0(PATH.TO.SIMULATION.DATA, i, "/", "c_bursts_analysis_", ANALYSIS,"_", EPSILON, ".dat"))
    simulated.bursts.per.pair[[i]] <- data$bursts.per.pair
  }

  # average simulated c-burst data
  matrix.simulated.bursts.per.pair <- do.call(rbind, simulated.bursts.per.pair)
  mean.simulated.bursts.per.pair <- colMeans(matrix.simulated.bursts.per.pair)
  median.simulated.bursts.per.pair <- colMedians(matrix.simulated.bursts.per.pair)

  pop.var <- function(x) var(x) * (length(x)-1) / length(x)
  pop.sd <- function(x) sqrt(pop.var(x))

  cat(c(mean(real.bursts.per.pair[real.bursts.per.pair != 0]), mean(median.simulated.bursts.per.pair[real.bursts.per.pair != 0]), pop.sd(real.bursts.per.pair[real.bursts.per.pair != 0]), pop.sd(median.simulated.bursts.per.pair[real.bursts.per.pair != 0]), median(real.bursts.per.pair[real.bursts.per.pair != 0]), median(median.simulated.bursts.per.pair[real.bursts.per.pair != 0]), "\n"), sep=", ")

  cat(c(mean(real.bursts.per.pair), mean(median.simulated.bursts.per.pair), pop.sd(real.bursts.per.pair), pop.sd(median.simulated.bursts.per.pair), median(real.bursts.per.pair), median(median.simulated.bursts.per.pair), "\n"), sep=", ")

  if(median) {
    m.simulated.bursts.per.pair=median.simulated.bursts.per.pair
  } else {
    m.simulated.bursts.per.pair=mean.simulated.bursts.per.pair
  }

  # perform test
  if(test=="t") {
    test <- t.test
    effect.size <- cohensD(x=real.bursts.per.pair, y=m.simulated.bursts.per.pair, method="paired")
  } else if(test=="MWU") {
    test <- wilcox.test
    #effect.size <- effsize::cliff.delta(d=real.bursts.per.pair, f=m.simulated.bursts.per.pair)$estimate
    grouped.data.vector.w0 = c(real.bursts.per.pair[real.bursts.per.pair != 0], m.simulated.bursts.per.pair[real.bursts.per.pair != 0])
    grouped.data.w0 = data.frame(x=real.bursts.per.pair[real.bursts.per.pair != 0], y=m.simulated.bursts.per.pair[real.bursts.per.pair !    = 0])
    effect.size.w0 <- wilcoxonPairedR(x=grouped.data.vector.w0, g=c(rep("real.bursts", nrow(grouped.data.w0)), rep("m.simulated.bursts", nrow(grouped.data.w0))))

    grouped.data.vector.i0 = c(real.bursts.per.pair, m.simulated.bursts.per.pair)
    grouped.data.i0 = data.frame(x=real.bursts.per.pair, y=m.simulated.bursts.per.pair)
    effect.size.i0 <- wilcoxonPairedR(x=grouped.data.vector.i0, g=c(rep("real.bursts", nrow(grouped.data.i0)), rep("m.simulated.bursts", nrow(grouped.data.i0))))
  }
  t.w0 <- test(x=real.bursts.per.pair[real.bursts.per.pair != 0], y=m.simulated.bursts.per.pair[real.bursts.per.pair != 0], alternative = "greater", paired=paired)
  t.i0 <- test(x=real.bursts.per.pair, y=m.simulated.bursts.per.pair, alternative = "greater", paired=paired)
  #  t <- shapiro.test(x=real.bursts.per.pair)

  return(list(t.w0,effect.size.w0,t.i0,effect.size.i0))
}

synchronicity.degree.t.test <- function(test="t"){

  # load real c-burst data
  load(paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", ANALYSIS,"_", EPSILON, ".dat"))
  real.sync.degrees <- data$sync.degrees.weighted

  # load simulated c-burst data
  simulated.sync.degrees <- vector("list", SIMULATIONS)

  for(i in c(1:SIMULATIONS)) {
    load(paste0(PATH.TO.SIMULATION.DATA, i, "/", "c_bursts_analysis_", ANALYSIS,"_", EPSILON, ".dat"))
    simulated.sync.degrees[[i]] <- data$sync.degrees.weighted
  }
  simulated.sync.degrees <- unlist(simulated.sync.degrees)

  pop.var <- function(x) var(x) * (length(x)-1) / length(x)
  pop.sd <- function(x) sqrt(pop.var(x))

  cat(c(mean(real.sync.degrees), mean(simulated.sync.degrees), pop.sd(real.sync.degrees), pop.sd(simulated.sync.degrees), median(real.sync.degrees), median(simulated.sync.degrees), "\n"), sep=", ")

  # perform test
  if(test=="t") {
    test <- t.test
    effect.size <- cohensD(x=real.sync.degrees, y=simulated.sync.degrees, method="pooled") #effsize::cohen.d(d=real.sync.degrees, f=simulated.sync.degrees, paired=FALSE)$estimate
  } else if(test=="MWU") {
    test <- wilcox.test
    effect.size <- effsize::cliff.delta(d=real.sync.degrees, f=simulated.sync.degrees)$estimate
  }
  t <- test(x=real.sync.degrees, y=simulated.sync.degrees, alternative = "greater", paired=FALSE)
  #  t <- shapiro.test(x=real.bursts.per.pair)

  return(list(t,effect.size))
}

code.growth.t.test <- function(test="t"){

  # load code-growth data
  load(paste0(PATH.TO.DATA, "/", "code_growth_metrics_", ANALYSIS,"_", EPSILON, ".dat"))
  non.synchronous <- calculations$non.synchronous.metrics
  synchronous <- calculations$synchronous.metrics

  if(test=="t"){
    test = t.test
    eff.size <- function(x, y) {
      return(cohensD(x=x, y=y))
    }
  } else {
    test = wilcox.test
    eff.size <- function(x, y) {
      return(cliff.delta(d=x, f=y)$estimate)
    }
  }

  t <- list()
  effect.size <- list()

  pop.var <- function(x) var(x) * (length(x)-1) / length(x)
  pop.sd <- function(x) sqrt(pop.var(x))

  cat("Growth: \n")
  cat(c(mean(synchronous$growth), mean(non.synchronous$growth), pop.sd(synchronous$growth), pop.sd(non.synchronous$growth), median(synchronous$growth), median(non.synchronous$growth), "\n"), sep=", ")
  cat("Effort: \n")
  cat(c(mean(synchronous$effort), mean(non.synchronous$effort), pop.sd(synchronous$effort), pop.sd(non.synchronous$effort), median(synchronous$effort), median(non.synchronous$effort), "\n"), sep=", ")

  # perform t-test on "added"
  t$added <- test(x=synchronous$added, y=non.synchronous$added, alternative = "greater", paired=FALSE)
  effect.size$added <- eff.size(x=synchronous$added, y=non.synchronous$added)

  # perform t-test on "deleted"
  t$deleted <- test(x=synchronous$deleted, y=non.synchronous$deleted, alternative = "less", paired=FALSE)
  effect.size$deleted <- eff.size(x=synchronous$deleted, y=non.synchronous$deleted)

  # perform t-test on "growth"
  t$growth <- test(x=synchronous$growth, y=non.synchronous$growth, alternative = "greater", paired=FALSE)
  effect.size$growth <- eff.size(x=synchronous$growth, y=non.synchronous$growth)

  # perform t-test on "effort"
  t$effort <- test(x=synchronous$effort, y=non.synchronous$effort, alternative = "less", paired=FALSE)
  effect.size$effort <- eff.size(x=synchronous$effort, y=non.synchronous$effort)

  # # perform t-test on "normlizedAdded"
  # t$normalizedAdded <- test(x=synchronous$normalizedAdded, y=non.synchronous$normalizedAdded, alternative = "greater", paired=FALSE)
  #
  # # perform t-test on "normalizedDelted"
  # t$normalizedDeleted <- test(x=synchronous$normalizedDeleted, y=non.synchronous$normalizedDeleted, alternative = "less", paired=FALSE)
  #
  # # perform t-test on "normalizedGrowth"
  # t$normalizedGrowth <- test(x=synchronous$normalizedGrowth, y=non.synchronous$normalizedGrowth, alternative = "greater", paired=FALSE)
  #
  # # perform t-test on "normalizedEffort"
  # t$normalizedEffort <- test(x=synchronous$normalizedEffort, y=non.synchronous$normalizedEffort, alternative = "less", paired=FALSE)

  return(list(t, effect.size))
}

### pairwise test returning W-values resp. t-values
# (adjusted copy of pairwise.wilcox.test)
pairwise.test.w = function (x, g, p.adjust.method = p.adjust.methods, paired = FALSE, test.method, ...) {

  p.adjust.method <- match.arg(p.adjust.method)
  DNAME <- paste(deparse(substitute(x)), "and", deparse(substitute(g)))
  g <- factor(g)
  METHOD <- if(identical(test.method,wilcox.test)) {
    if (paired) "Wilcoxon signed rank" else "Wilcoxon rank sum"
  } else if(identical(test.method,t.test)){
    if (paired) "Paired t-test" else "Welch Two Sample t-test"
  }
  compare.levels.w <- function(i, j) {
    xi <- x[as.integer(g) == i]
    xj <- x[as.integer(g) == j]
    test.method(xi, xj, paired = paired, ...)$statistic
  }

  WVAL <- pairwise.table(compare.levels.w, levels(g), p.adjust.method)
  ans <- list(method = METHOD, data.name = DNAME, test.value = WVAL)
  ans
}

### pairwise test returning df values
# (adjusted copy of pairwise.wilcox.test)
pairwise.test.df = function (x, g, p.adjust.method = p.adjust.methods, paired = FALSE, test.method, ...) {

  if(identical(test.method, wilcox.test)) {
    return()
  }

  p.adjust.method <- match.arg(p.adjust.method)
  DNAME <- paste(deparse(substitute(x)), "and", deparse(substitute(g)))
  g <- factor(g)
  METHOD <- if(identical(test.method,wilcox.test)) {
    if (paired) "Wilcoxon signed rank" else "Wilcoxon rank sum"
  } else if(identical(test.method,t.test)){
    if (paired) "Paired t-test" else "Welch Two Sample t-test"
  }
  compare.levels.df <- function(i, j) {
    xi <- x[as.integer(g) == i]
    xj <- x[as.integer(g) == j]
    test.method(xi, xj, paired = paired, ...)$parameter
  }

  DFVAL <- pairwise.table(compare.levels.df, levels(g), p.adjust.method)
  ans <- list(method = METHOD, data.name = DNAME, test.value = DFVAL)
  ans
}

### pairwise test returning adjusted p-values
# (adjusted copy of pairwise.wilcox.test)
pairwise.test.p = function (x, g, p.adjust.method = p.adjust.methods, paired = FALSE, test.method, ...) {

  p.adjust.method <- match.arg(p.adjust.method)
  DNAME <- paste(deparse(substitute(x)), "and", deparse(substitute(g)))
  g <- factor(g)
  METHOD <- if(identical(test.method,wilcox.test)) {
    if (paired) "Wilcoxon signed rank" else "Wilcoxon rank sum"
  } else if(identical(test.method,t.test)){
    if (paired) "Paired t-test" else "Welch Two Sample t-test"
  }
  compare.levels.p <- function(i, j) {
    xi <- x[as.integer(g) == i]
    xj <- x[as.integer(g) == j]
    test.method(xi, xj, paired = paired, ...)$p.value
  }
  PVAL <- pairwise.table(compare.levels.p, levels(g), p.adjust.method)
  ans <- list(method = METHOD, data.name = DNAME, p.value = PVAL,
              p.adjust.method = p.adjust.method)
  #class(ans) <- "pairwise.htest"
  ans
}

compare.synchronous.development.test <- function(test.type="t", input="bursts.per.pair") {
  loginfo("COMPARE FILES, FEATURES AND FUNCTIONS")

  files <- "files"
  features <- "features"
  functions <- "functions"

  # load real c-burst data files
  load(paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", files, "_", EPSILON, ".dat"))
  real.bursts.data.files <- sort(data[[input]])

  # load real c-burst data features
  load(paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", features, "_", EPSILON, ".dat"))
  real.bursts.data.features <- sort(data[[input]])

  # load real c-burst data functions
  load(paste0(PATH.TO.DATA, "/", "c_bursts_analysis_", functions ,"_", EPSILON, ".dat"))
  real.bursts.data.functions <- sort(data[[input]])

  if(test.type=="t") {
    test <- t.test
  } else if(test.type=="MWU") {
    test <- wilcox.test
  }

  x <- c(real.bursts.data.functions,
         real.bursts.data.files,
         real.bursts.data.features)

  g <- c(rep("functions",length(real.bursts.data.functions)),
      rep("files",length(real.bursts.data.files)),
      rep("features",length(real.bursts.data.features)))

  alternative <- "less"

  # perform test
  w = pairwise.test.w(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  df = pairwise.test.df(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                        paired = FALSE, exact = FALSE, test.method = test)

  p = pairwise.test.p(x=x, g=g, p.adjust.method = "fdr", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  # construct pairs with significant results
  significants = which(p$p.value < 0.05, arr.ind = TRUE)
  significants.count = nrow(significants)

  # adjust significance level
  if(significants.count > 0) {
    clevel = 1 - ((1 - 0.95) / significants.count)
  } else {
    clevel = 0.95
  }

  if(test.type=="t") {
    effect.size <- effsize::cohen.d(d=real.bursts.data.features, f=real.bursts.data.files, conf.level = clevel)$estimate
  } else if(test.type=="MWU") {
    effect.size <- effsize::cliff.delta(d=real.bursts.data.features, f=real.bursts.data.files, conf.level = clevel)$estimate
  }

  return(list(w,df,p,effect.size))
}

compare.code.growth.test <- function(test.type="t", metric="added") {
  loginfo("COMPARE FILES, FEATURES AND FUNCTIONS")

  files <- "files"
  features <- "features"
  functions <- "functions"

  # load code-growth data
  #load(paste0(PATH.TO.DATA, "/", "code_growth_metrics", ANALYSIS,"_", EPSILON, ".dat"))
  #non.synchronous <- calculations$non.synchronous.metrics
  #synchronous <- calculations$synchronous.metrics

  # load real code-growth data files
  load(paste0(PATH.TO.DATA, "/", "code_growth_metrics_", files,"_", EPSILON, ".dat"))
  non.synchronous.commits.files <- calculations$non.synchronous.metrics[[metric]]
  synchronous.commits.files <- calculations$synchronous.metrics[[metric]]

  # load real code-growth data features
  load(paste0(PATH.TO.DATA, "/", "code_growth_metrics_", features,"_", EPSILON, ".dat"))
  non.synchronous.commits.features <- calculations$non.synchronous.metrics[[metric]]
  synchronous.commits.features <- calculations$synchronous.metrics[[metric]]

  # load real code-growth data functions
  load(paste0(PATH.TO.DATA, "/", "code_growth_metrics_", functions,"_", EPSILON, ".dat"))
  non.synchronous.commits.functions <- calculations$non.synchronous.metrics[[metric]]
  synchronous.commits.functions <- calculations$synchronous.metrics[[metric]]

  if(test.type=="t") {
    test <- t.test
  } else if(test.type=="MWU") {
    test <- wilcox.test
  }

  x <- c(synchronous.commits.functions,
         synchronous.commits.files,
         synchronous.commits.features)

  g <- c(rep("functions",length(synchronous.commits.functions)),
         rep("files",length(synchronous.commits.files)),
         rep("features",length(synchronous.commits.features)))

  alternative <- "less"

  # perform test
  w = pairwise.test.w(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  df = pairwise.test.df(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                        paired = FALSE, exact = FALSE, test.method = test)

  p = pairwise.test.p(x=x, g=g, p.adjust.method = "fdr", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  # construct pairs with significant results
  significants = which(p$p.value < 0.05, arr.ind = TRUE)
  significants.count = nrow(significants)

  # adjust significance level
  if(significants.count > 0) {
    clevel = 1 - ((1 - 0.95) / significants.count)
  } else {
    clevel = 0.95
  }

  if(test.type=="t") {
    effect.size <- effsize::cohen.d(d=synchronous.commits.features, f=synchronous.commits.files, conf.level = clevel)$estimate
  } else if(test.type=="MWU") {
    effect.size <- effsize::cliff.delta(d=synchronous.commits.features, f=synchronous.commits.files, conf.level = clevel)$estimate
  }

  return(list(w,df,p,effect.size))
}

c.e.dtw.correlation.t.test <- function(test="t") {
  # FIXME
  bandwindowsize = 2 * EPSILON

  # load real data
  load(paste0(PATH.TO.DATA, "/", "smoothed_curves_", threads, ANALYSIS ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON,
              "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
  real.dtw <- unlist(data$dtw.distances)
  print(paste("Print real dtw mean:" ,mean(real.dtw)))

  ## load simulated data
  #load(paste0(PATH.TO.DATA, "/", "simulation1/simulated_smoothed_curves_", threads, ANALYSIS ,"_", EPSILON, "_bandWindowSize", bandwindowsize, ".dat"))
  #simulated.dtw <- unlist(data$dtw.distances)
  #print(paste("Print simulated dtw mean:" ,mean(simulated.dtw)))

# load simulated data
  simulated.dtw <- vector("list", SIMULATIONS)

  for(i in c(1:SIMULATIONS)) {
    file <- paste0(PATH.TO.DATA, "/", "simulation", i, "/simulated_smoothed_curves_", threads, ANALYSIS ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat")
    if(file.exists(file)) {
      load(paste0(PATH.TO.DATA, "/", "simulation", i, "/simulated_smoothed_curves_", threads, ANALYSIS ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
    } else {
      load(paste0(PATH.TO.DATA, "/", "simulation", i, "/simulated_smoothed_curves_", threads, ANALYSIS ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, ".dat"))
    }
    simulated.dtw[[i]] <- unlist(data$dtw.distances)
  }
  print(paste("Print simulated dtw mean:", mean(unlist(simulated.dtw))))

  simulated.all.dtw <- unlist(simulated.dtw)

  pop.var <- function(x) var(x) * (length(x)-1) / length(x)
  pop.sd <- function(x) sqrt(pop.var(x))

  cat("Real vs Simulated: \n")
  cat(c(mean(real.dtw), mean(simulated.all.dtw), pop.sd(real.dtw), pop.sd(simulated.all.dtw), median(real.dtw), median(simulated.all.dtw), "\n"), sep=", ")

  # perform test
  if(test=="t") {
    test <- t.test
    effect.size <- cohensD(x=real.dtw, y=simulated.all.dtw, method="pooled") #
    #effect.size <- effsize::cohen.d(d=real.dtw, f=simulated.dtw, paired=FALSE)$estimate
  } else if(test=="MWU") {
    test <- wilcox.test
    effect.size <- effsize::cliff.delta(d=real.dtw, f=simulated.all.dtw)$estimate
  }
  t <- test(x=real.dtw, y=simulated.all.dtw, alternative = "less", paired=FALSE)

  #t <- shapiro.test(x=real.dtw)

  return(list(t,effect.size))
}

compare.dtw.test <- function(test.type="mtw") {
  loginfo("COMPARE FILES, FEATURES AND FUNCTIONS")

  bandwindowsize = 2 * EPSILON

  files <- "files"
  features <- "features"
  functions <- "functions"

  # load real data files
  load(paste0(PATH.TO.DATA, "/", "smoothed_curves_", threads, files ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
  real.dtw.data.files <- unlist(data$dtw.distances)

  # load real data features
  load(paste0(PATH.TO.DATA, "/", "smoothed_curves_", threads, features ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
  real.dtw.data.features <- unlist(data$dtw.distances)

  # load real data functions
  load(paste0(PATH.TO.DATA, "/", "smoothed_curves_", threads, functions ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
  real.dtw.data.functions <- unlist(data$dtw.distances)

  if(test.type=="t") {
    test <- t.test
  } else if(test.type=="MWU") {
    test <- wilcox.test
  }

  x <- c(real.dtw.data.functions,
         real.dtw.data.files,
         real.dtw.data.features)

  g <- c(rep("functions",length(real.dtw.data.functions)),
         rep("files",length(real.dtw.data.files)),
         rep("features",length(real.dtw.data.features)))

  alternative <- "greater"

  # perform test
  w = pairwise.test.w(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  df = pairwise.test.df(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                        paired = FALSE, exact = FALSE, test.method = test)

  p = pairwise.test.p(x=x, g=g, p.adjust.method = "fdr", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  # construct pairs with significant results
  significants = which(p$p.value < 0.05, arr.ind = TRUE)
  significants.count = nrow(significants)

  # adjust significance level
  if(significants.count > 0) {
    clevel = 1 - ((1 - 0.95) / significants.count)
  } else {
    clevel = 0.95
  }

  if(test.type=="t") {
    effect.size <- effsize::cohen.d(d=real.dtw.data.features, f=real.dtw.data.files, conf.level = clevel)$estimate
  } else if(test.type=="MWU") {
    effect.size <- effsize::cliff.delta(d=real.dtw.data.features, f=real.dtw.data.files, conf.level = clevel)$estimate
  }

  return(list(w,df,p,effect.size))
}

compare.dtw.test.2 <- function(test.type="mtw") {
  loginfo("COMPARE THREADS AND MESSAGES")

  bandwindowsize = 2 * EPSILON

  thread.based  <- "thread_"
  no.threads    <- ""

  # load real data non-thread-based
  load(paste0(PATH.TO.DATA, "/", "smoothed_curves_", no.threads, ANALYSIS ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
  real.dtw.data.no.threads <- unlist(data$dtw.distances)

  # load real data thread-based
  load(paste0(PATH.TO.DATA, "/", "smoothed_curves_", thread.based, ANALYSIS ,"_", COMMIT.MESSAGES, incl_less_5, EPSILON, "_SigmaAsBandWindow_bandWindowSize", bandwindowsize, ".dat"))
  real.dtw.data.threads <- unlist(data$dtw.distances)

  if(test.type=="t") {
    test <- t.test
  } else if(test.type=="MWU") {
    test <- wilcox.test
  }

  x <- c(real.dtw.data.no.threads,
         real.dtw.data.threads)

  g <- c(rep("no.threads",length(real.dtw.data.no.threads)),
         rep("thread.based",length(real.dtw.data.threads)))

  alternative <- "less"

  # perform test
  w = pairwise.test.w(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  df = pairwise.test.df(x=x, g=g, p.adjust.method = "none", alternative=alternative,
                        paired = FALSE, exact = FALSE, test.method = test)

  p = pairwise.test.p(x=x, g=g, p.adjust.method = "fdr", alternative=alternative,
                      paired = FALSE, exact = FALSE, test.method = test)

  # construct pairs with significant results
  significants = which(p$p.value < 0.05, arr.ind = TRUE)
  significants.count = nrow(significants)

  # adjust significance level
  if(significants.count > 0) {
    clevel = 1 - ((1 - 0.95) / significants.count)
  } else {
    clevel = 0.95
  }

  if(test.type=="t") {
    effect.size <- effsize::cohen.d(d=real.dtw.data.no.threads, f=real.dtw.data.threads, conf.level = clevel)$estimate
  } else if(test.type=="MWU") {
    effect.size <- effsize::cliff.delta(d=real.dtw.data.no.threads, f=real.dtw.data.threads, conf.level = clevel)$estimate
  }

  return(list(w,df,p,effect.size))
}


print("synchronous development real vs. simulated")
print("##########################################")
print("  ")
synchronous.development.t.test(paired=TRUE, median=TRUE, test="MWU")

print("synchronicity degree real vs. simulated")
print("##########################################")
print("  ")
synchronicity.degree.t.test(test="MWU")

print("codegrowth synchronous vs. non-synchronous")
print("##########################################")
print("  ")
code.growth.t.test(test="MWU")

print("synchronous development features vs files vs functions")
print("##########################################")
print("  ")
compare.synchronous.development.test(test.type="MWU", input="bursts.per.pair")

print("synchronicity degree features vs files vs functions")
print("##########################################")
print("  ")
compare.synchronous.development.test(test.type="MWU", input="sync.degrees.weighted")

print("code added features vs files vs functions")
print("##########################################")
print("  ")
compare.code.growth.test(test.type="MWU", metric="added")

print("code deleted features vs files vs functions")
print("##########################################")
print("  ")
compare.code.growth.test(test.type="MWU", metric="deleted")

print("code growth features vs files vs functions")
print("##########################################")
print("  ")
compare.code.growth.test(test.type="MWU", metric="growth")

print("code effort features vs files vs functions")
print("##########################################")
print("  ")
compare.code.growth.test(test.type="MWU", metric="effort")

print("DTW distances C-E real vs simulated")
print("##########################################")
print("  ")
#c.e.dtw.correlation.t.test(test="MWU")

print("DTW distances features vs files vs functions")
print("##########################################")
print("  ")
compare.dtw.test(test.type="MWU")

print("DTW distances threads vs messages")
print("##########################################")
print("  ")
compare.dtw.test.2(test.type="MWU")
