## Copyright (c) Thomas Bock, 2016-2018
## bockthom@fim.uni-passau.de
##
## Copyright (c) Claus Hunsen, 2016
## hunsen@fim.uni-passau.de
##
## Distributed under terms of the MIT license.

## libraries
suppressWarnings(library(plyr)) # tools for splitting, applying and combining data
suppressWarnings(library(lubridate)) # tools for handling dates
suppressWarnings(library(logging)) # logging

## / / / / / / / / / / / / / /
## Utility functions
## / / / / / / / / / / / / / /

read.list.from.file = function(file, sep = ";") {

  # read data.frame from disk (as expected from save.list.to.file)
  d = read.table(file, header = FALSE, sep = sep, strip.white = TRUE)
}

read.commit.data <- function(path.to.commit.list, artefact.type) {
  commit.list <- path.to.commit.list

  commits <- read.list.from.file(commit.list)
  colnames(commits) <- c("commitId", "authorDate", "author", "authorEmail", "commitHash",
                         "changedFiles", "addedLines", "deletedLines", "diffSize", "file", "artefact", "artefactType", "size")

  if(artefact.type == "feature") {
    commits <- commits[commits[["artefactType"]] == "Feature",]
    rownames(commits) <- seq(length=nrow(commits))
  } else if(artefact.type == "proximity") {
    commits[["artefact"]] <- paste(commits[["file"]], commits[["artefact"]], sep="::")
  }
  return(commits)
}

read.email.data <- function(path.to.email.list) {
  email.list <- path.to.email.list

  emails <- read.list.from.file(email.list)
  colnames(emails) <- c("author", "authorEmail", "messageId", "creationDate", "creationDateOffset", "subject", "threadId")

  emails <- na.omit(emails[emails["creationDate"] != "None",])
  emails <- na.omit(emails[emails["creationDate"] != "",])
  return(emails)
}

read.added.deleted.lines.data <- function(path.to.file) {
  # data extracted by following command:
  # for c in $(git rev-list --all); do git show --numstat $c | grep '^[[:digit:]]' | awk -v env_var="$c" ' { OFS=";"; print env_var,$3,$1,$2 >> "path.to.file";}'; done
  added.deleted.lines.data <- read.list.from.file(path.to.file)
  colnames(added.deleted.lines.data) <- c("commitHash", "file", "addedLines", "deletedLines")
  return(added.deleted.lines.data)
}

read.mboxparsing.data <- function(path.to.mboxparsing.list, artefact.type, commit.data = NULL) {
  mboxparsing.list <- path.to.mboxparsing.list

  mboxparsing <- read.list.from.file(mboxparsing.list)
  colnames(mboxparsing) <- c("file", "artefact", "messageId")

  if(artefact.type == "feature" && !is.null(commit.data)) {
    features <- commit.data[["artefact"]]
    mboxparsing <- mboxparsing[mboxparsing[["artefact"]] %in% features,]
    rownames(mboxparsing) <- seq(length=nrow(mboxparsing))
  } else if(artefact.type == "function") {
    mboxparsing[["artefact"]] <- paste(mboxparsing[["file"]], mboxparsing[["artefact"]], sep="::")
  } else if(artefact.type == "file") {
    mboxparsing[["artefact"]] <- mboxparsing[["file"]]
  }
  return(mboxparsing)
}

read.commit.messages.data <- function(path.to.commit.messages.list, filter.for.first.line = FALSE) {

  commit.messages <- read.list.from.file(path.to.commit.messages.list)
  colnames(commit.messages) <- c("commitId", "commitHash", "commitMessage")

  # throw away commitId
  commit.messages <- unique(commit.messages[,c("commitHash", "commitMessage")])

  if(filter.for.first.line){
    commit.messages[["commitMessage"]] <- sapply(commit.messages[["commitMessage"]], FUN=function(message){strsplit(message, split="    ", fixed=TRUE)[[1]][2]})
  }
  return(commit.messages)
}

get.pair.no <- function(developer1, developer2, pairs.of.developers.data.frame) {

  # get the key for a pair of developers, i.e., geth the rowname of the pair in the given data frame
  return(as.numeric(rownames(pairs.of.developers.data.frame[(pairs.of.developers.data.frame[,1]==developer1 & pairs.of.developers.data.frame[,2]==developer2),])))
}

get.pairs.for.one.developer <- function(developer1, pairs.of.developers.data.frame) {
  return(pairs.of.developers.data.frame[(pairs.of.developers.data.frame[,1]==developer1 | pairs.of.developers.data.frame[,2]==developer1),])
}

filter.data.by.date.interval <- function(data, date.column, end.date, start.date) {

  # if start or end dates are not specified, get minimal resp. maximal date of the data
  if(is.null(start.date)) {
    start.date <- min(data[,date.column])
  }
  if(is.null(end.date)) {
    end.date <- max(data[,date.column])
  }

  # disregard all data before and after certain start and end dates regarding the column 'date.column'
  return(data[ymd_hms(data[,date.column])<= ymd_hms(end.date) & ymd_hms(data[,date.column])>=ymd_hms(start.date),])
}

determine.length.of.interval.in.days <- function(interval) {
  return(as.numeric(as.duration(interval))/(3600*24.0)) # convert seconds to days
}

determine.length.of.interval.in.seconds <- function(interval) {
  return(as.numeric(as.duration(interval))) # in seconds
}

get.artefact.column <- function(analyzeFiles) {
  if(analyzeFiles) {
    return("file")
  } else {
    return("artefact")
  }
}

get.emails.by.author <- function(author, emails.per.author, authors) {
  return(emails.per.author[[(which(authors==author))]])
}

get.emails.of.author <- function(author, emails) {
  return(emails[emails[["author"]]==author,])
}

get.commits.of.author <- function(author, commits) {
  return(commits[commits[["author"]]==author,])
}

get.authors.of.commits <- function(commits){
  return(as.vector(unique(commits[["author"]])))
}

filter.commit.authors <- function(emails, commits) {
  # consider only authors which are also authors of a commit
  emails <- emails[emails[["author"]] %in% get.authors.of.commits(commits),]
  rownames(emails) <- seq(length=nrow(emails))
  return(emails)
}

remove.file.level.commits <- function(commits) {
  commits <- commits[commits[["artefact"]] != "File_Level",]
  rownames(commits) <- seq(length=nrow(commits))
  return(commits)
}

remove.base.feature.commits <- function(commits) {
  commits <- commits[commits[["artefact"]] != "Base_Feature",]
  rownames(commits) <- seq(length=nrow(commits))
  return(commits)
}

get.number.of.bursts.overall <- function(list.of.bursts) {

  # get overall number of bursts
  number.bursts.overall <- 0
  for(i in c(1:length(list.of.bursts))) {
    if(!is.null(nrow(list.of.bursts[[i]]))){
      number.bursts.overall <- number.bursts.overall + nrow(list.of.bursts[[i]])
    }
  }
  return(number.bursts.overall)
}

get.lengths.of.bursts <- function(pairs.of.authors, pairs.of.authors.data.frame, list.of.bursts) {

  # iterate over all bursts for each pair to determine the lengths of bursts
  length.of.intervals <- mclapply(pairs.of.authors, mc.cores=CORES, FUN=function(pair){
    length.of.intervals <- c()
    pair.no <- get.pair.no(pair[1], pair[2], pairs.of.authors.data.frame)
    bursts.data.frame <- list.of.bursts[[pair.no]]

    # Check if there is no burst for the current author pair
    if(length(bursts.data.frame)>0) {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(interval in bursts) {
        ival <- interval(interval[1], interval[2])

        # Add length of interval to the list
        length <- determine.length.of.interval.in.days(ival)
        length.of.intervals <- c(length.of.intervals, length[1])
      }
    }
    return(length.of.intervals)
  })
  length.of.intervals <- unlist(length.of.intervals)
  return(length.of.intervals)
}

get.activities.of.bursts <- function(pairs.of.authors, pairs.of.authors.data.frame, list.of.bursts, type, raw.data, raw.data.timestamp) {

  # iterate over all bursts for each pair to determine the number of activities per burst
  number.of.activities <- mclapply(pairs.of.authors, mc.cores=CORES, FUN=function(pair){
    number.of.activities <- c()
    pair.no <- get.pair.no(pair[1], pair[2], pairs.of.authors.data.frame)
    bursts.data.frame <- list.of.bursts[[pair.no]]

    if(type=="commit") {
      activities1 <- get.commits.of.author(pair[1], raw.data)
      activities2 <- get.commits.of.author(pair[2], raw.data)
    } else if(type=="email") {
      activities1 <- get.emails.of.author(pair[1], raw.data)
      activities2 <- get.emails.of.author(pair[2], raw.data)
    } else {
      stop("Wrong analysis type")
    }

    # Check if there is no burst for the current author pair
    if(length(bursts.data.frame)>0) {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      for(interval in bursts) {
        burst.activities1 <- activities1[(activities1[[raw.data.timestamp]] >= interval[1] & activities1[[raw.data.timestamp]]<= interval[2]), raw.data.timestamp]
        burst.activities2 <- activities2[(activities2[[raw.data.timestamp]] >= interval[1] & activities2[[raw.data.timestamp]] <= interval[2]), raw.data.timestamp]
        burst.activities1.count <- length(unique(burst.activities1))
        burst.activities2.count <- length(unique(burst.activities2))

        # Add number of burst-activities to list
        number.of.activities <- c(number.of.activities, burst.activities1.count+burst.activities2.count)
      }
    }
    return(number.of.activities)
  })
  number.of.activities <- unlist(number.of.activities)
  return(number.of.activities)
}

get.number.of.bursts.per.pair <- function(pairs.of.authors, pairs.of.authors.data.frame, list.of.bursts) {

  # iterate over all pairs to determine the number of bursts per pair
  number.of.intervals <- mclapply(pairs.of.authors, mc.cores=CORES, FUN=function(pair){
    number.of.intervals <- c()
    pair.no <- get.pair.no(pair[1], pair[2], pairs.of.authors.data.frame)
    bursts.data.frame <- list.of.bursts[[pair.no]]

    # Check if there is no burst for the current author pair
    if(length(bursts.data.frame)==0) {
      number.of.intervals <- c(number.of.intervals, 0)
    } else {
      bursts <- as.list(as.data.frame(t(bursts.data.frame)))

      # Determine number of bursts for the current pair and add number to the list
      number.of.intervals <- c(number.of.intervals, length(bursts))
    }
    return(number.of.intervals)
  })
  number.of.intervals <- unlist(number.of.intervals)
  return(number.of.intervals)
}

fix.midnight <- function(timestamp) {
  if(!grepl(":", timestamp)) {
    timestamp <- paste(timestamp, "00:00:00")
  }
  return(timestamp)
}
